Require Import haskell.
Require Import haskell_well_formed.
Require Import language_requirements.
Require Import maps.
Require Import models.
Require Import types.

Module Haskell_Language_Requirements <: LANGUAGE_REQUIREMENTS.
  Module Typ := Haskell_Well_Formed.
  Import Typ.
  Import L.

  Inductive program_order : effect val -> effect val -> Prop :=
  | program_order_same : forall t a a',
    program_order (Some (t, a)) (Some (t, a'))
  | program_order_spawn : forall t t' a,
    program_order (Some (t, spawn _ t')) (Some (t', a)).

  (* This is one of those proofs that's just ugly and annoying. *)
  Lemma program_reordering : forall P1 P2 P3 e e',
    program_step P1 e P2 ->
    program_step P2 e' P3 ->
    ~ program_order e e' ->
    exists P2', exists P3',
      eq_program_state P3 P3'
      /\ program_step P1 e' P2'
      /\ program_step P2' e P3'.
  Proof with eauto.
    intros P1 P2 P3 e e' Step1 Step2 NotOrder.
    inversion Step1 as [ ? t1 | ? t1 ? ? ? t2 ? FreshTID2 ]; subst;
      inversion Step2 as [ ? t3 | ? t3 ? ? ? t4 ? FreshTID4 ]; subst;
        (assert (t1 <> t3); [ intro; subst;
          apply NotOrder; apply program_order_same | ]).

    (* non-spawn/non-spawn *)
    esplit. esplit.
    rewrite and_comm.
    split. split.
    eapply step_thread...
    rewrite <- VarMapFacts.add_neq_mapsto_iff...
    eapply step_thread...
    rewrite VarMapFacts.add_neq_mapsto_iff...
    apply VarMapFactsEx2.Equal_add_add...

    (* non-spawn/spawn *)
    rewrite VarMapFactsEx.add_not_in_iff in FreshTID4.
    destruct FreshTID4 as [ ? FreshTID4 ].
    assert (t3 <> t4).
    intro. subst.
    contradiction FreshTID4.
    rewrite <- VarMapFacts.add_neq_in_iff...
    esplit. esplit.
    rewrite and_comm.
    split. split.
    eapply step_spawn...
    rewrite <- VarMapFacts.add_neq_mapsto_iff...
    eapply step_thread...
    rewrite VarMapFacts.add_neq_mapsto_iff...
    rewrite VarMapFacts.add_neq_mapsto_iff...
    apply VarMapFacts.Equal_sym.
    eapply VarMapFacts.Equal_trans.
    apply VarMapFactsEx2.Equal_add_add...
    apply VarMapProperties.F.add_m...
    apply VarMapFactsEx2.Equal_add_add...

    (* non-spawn/spawn *)
    assert (t2 <> t3).
    intro. subst.
    apply NotOrder.
    apply program_order_spawn.
    esplit. esplit.
    rewrite and_comm.
    split. split.
    eapply step_thread...
    rewrite <- VarMapFacts.add_neq_mapsto_iff.
    rewrite <- VarMapFacts.add_neq_mapsto_iff...
    auto.
    eapply step_spawn...
    rewrite VarMapFacts.add_neq_mapsto_iff...
    rewrite VarMapFactsEx.add_not_in_iff...
    eapply VarMapFacts.Equal_trans.
    apply VarMapFactsEx2.Equal_add_add...
    apply VarMapProperties.F.add_m...
    apply VarMapFactsEx2.Equal_add_add...

    (* spawn/spawn *)
    assert (t3 <> t4).
    intro. subst.
    contradiction FreshTID4...
    assert (t1 <> t2).
    intro. subst.
    contradiction FreshTID2...
    rewrite VarMapFactsEx.add_not_in_iff in FreshTID4.
    destruct FreshTID4 as [ ? FreshTID4 ].
    rewrite VarMapFactsEx.add_not_in_iff in FreshTID4.
    destruct FreshTID4 as [ ? FreshTID4 ]...
    assert (t2 <> t3).
    intro. subst.
    apply NotOrder.
    apply program_order_spawn.
    esplit. esplit.
    rewrite and_comm.
    split. split.
    eapply step_spawn...
    rewrite <- VarMapFacts.add_neq_mapsto_iff.
    rewrite <- VarMapFacts.add_neq_mapsto_iff...
    auto.
    eapply step_spawn...
    rewrite VarMapFacts.add_neq_mapsto_iff...
    rewrite VarMapFacts.add_neq_mapsto_iff...
    rewrite VarMapFactsEx.add_not_in_iff.
    rewrite VarMapFactsEx.add_not_in_iff...
    eapply VarMapFacts.Equal_trans.
    apply VarMapFactsEx2.Equal_add_add_add...
    apply VarMapProperties.F.add_m...
    eapply VarMapFacts.Equal_trans.
    apply VarMapFactsEx2.Equal_add_add_add...
    apply VarMapProperties.F.add_m...
    apply VarMapFacts.Equal_refl...
  Qed.

  Hint Constructors reduce.

  Lemma value_independent_reads_reduce : forall M1 d M2 M1' M2',
    reduce M1 (read d M2) M1' None ->
    exists M1'',
      reduce M1 (read d M2') M1'' None.
  Proof with eauto.
    intros M1 d M2 M1' M2' Reduce.
    remember (read d M2) as a.
    remember (@None trm) as M_opt.
    induction Reduce; try discriminate;
      try (forwards [ ? ? ] : IHReduce (>>> __ __); eauto);
        esplit...
    inversion Heqa. subst...
  Qed.

  Lemma value_independent_reads : forall P t d M P' M',
    program_step P (Some (t, read d M)) P' ->
    exists P', program_step P (Some (t, read d M')) P'.
  Proof with eauto.
    intros P t d M P' M' Step.
    inversion Step; subst.
    forwards [ ? ? ] : value_independent_reads_reduce...
    esplit. eapply step_thread...
  Qed.

  Import Haskell_Well_Formed_Preservation.

  Lemma well_formed_read_reduce : forall PT ST M1 d M2 M1',
    typing_trm (PT, ST) M1 ->
    reduce M1 (read d M2) M1' None ->
    VarMap.In d ST.
  Proof with eauto.
    intros PT ST M1 d M2 M1' TypingTrm Reduce.
    remember (read d M2) as a.
    remember (@None trm) as M_opt.
    induction Reduce; try discriminate;
      inversion_typing_trm TypingTrm TypingDVar x...
    inversion Heqa. subst.
    inversion TypingDVar...
  Qed.

  Lemma well_formed_read : forall PT ST P t d M P',
    Typ.typing_prgm (PT, ST) P ->
    program_step P (Some (t, read d M)) P' ->
    VarMap.In d ST.
  Proof with eauto.
    intros PT ST P t d M P' TypingPrgm Step.
    inversion TypingPrgm; subst.
    inversion Step; subst.
    eauto using well_formed_read_reduce.
  Qed.

  Lemma well_formed_write : forall PT ST P t d M P',
    Typ.typing_prgm (PT, ST) P ->
    program_step P (Some (t, write d M)) P' ->
    VarMap.In d ST.
  Proof with eauto.
    intros PT ST P t d M P' TypingPrgm Step.
    forwards [ ? [ TypingAction ? ] ] : type_preservation_program_step_mutual...
    apply well_typed_heap_action_write.
    inversion TypingAction...
  Qed.
End Haskell_Language_Requirements.
