Require Export haskell.
Require Export haskell_language_requirements.
Require Export haskell_well_formed.
Require Export heaps.
Require Export indexes.
Require Export inverted_lazy.
Require Export inverted_weak.
Require Export language_requirements.
Require Export maps.
Require Export model_facts.
Require Export models.
Require Export strong.
Require Export strong_lazy.
Require Export typed_heaps.
Require Export types.
Require Export weak.

Set Implicit Arguments.

Module L := Haskell.
Module LangFacts := LanguageModelFacts L.

Declare Module Heap : HEAP L.

Module Strong := Strong L Heap.
Module StrongFacts := MemoryModelFacts L Strong.

Module Lazy := StrongLazy L Heap.
Module LazyFacts := MemoryModelFacts L Lazy.

Module Weak := Weak L Heap.
Module WeakFacts := MemoryModelFacts L Weak.

Module ILazy := InvertedLazy L Heap.
Module ILazyFacts := MemoryModelFacts L ILazy.

Module IWeak := InvertedWeak L Heap.
Module IWeakFacts := MemoryModelFacts L IWeak.

Declare Module StrongSys : SYSTEM_MODEL L Strong.
Module StrongSysFacts := ModelFacts L Strong StrongSys.

Declare Module LazySys : SYSTEM_MODEL L Lazy.
Module LazySysFacts := ModelFacts L Lazy LazySys.

Declare Module WeakSys : SYSTEM_MODEL L Weak.
Module WeakSysFacts := ModelFacts L Weak WeakSys.

Declare Module ILazySys : SYSTEM_MODEL L ILazy.
Module ILazySysFacts := ModelFacts L ILazy ILazySys.

Declare Module IWeakSys : SYSTEM_MODEL L IWeak.
Module IWeakSysFacts := ModelFacts L IWeak IWeakSys.

Module Typ := Haskell_Well_Formed.

Module PrgmPres := Haskell_Well_Formed_Preservation.

Module LangReqs := Haskell_Language_Requirements.
