Require Import init_modules.
Require Import inverted_to_strong_map_heap.

Lemma step_pure : forall H3 H1 H3' t,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, pure _)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, pure _)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r v v' Active RgnFresh Fresh Read
    Empty.
  esplit. split.
  apply Strong.step_pure...
  auto.
Qed.

Lemma step_newRgn : forall H3 H1 H3' t r,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, newRgn _ r)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, newRgn _ r)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r' v v' Active RgnFresh Fresh Read
    Empty.
  esplit. split.
  apply Strong.step_newRgn...
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_alloc : forall H3 H1 H3' t l r v,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, alloc l r v)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, alloc l r v)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t l r v MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l' s r' v' v'' Active RgnFresh Fresh Read
    Empty.
  esplit. split.
  apply Strong.step_alloc.
  rewrite VarMapFacts.not_find_in_iff.
  rewrite VarMapFacts.not_find_in_iff in Fresh.
  specialize (MapStore l).
  rewrite Fresh in MapStore...
  apply map_heap_alloc...
Qed.

Lemma step_read : forall H3 H1 H3' t l v,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, read l v)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, read l v)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t l v MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  specialize (MapStore l).
  inversion_inverted_lazy_step Step t' l' s r v' v'' Active RgnFresh Fresh Read
    Empty; subst.

  (* tx_read1 *)
  esplit. split.
  eapply Strong.step_read.
  apply Heap.strong_active_running.
  eapply map_heap_read...
  apply map_heap_tx_read1...

  (* tx_read2 *)
  esplit. split...
  eapply Strong.step_read.
  apply Heap.strong_active_running.
  eapply map_heap_read...

  (* read *)
  esplit. split...
  eapply Strong.step_read.
  apply Heap.strong_active_none.
  eapply map_heap_read...
Qed.

Lemma step_write : forall H3 H1 H3' t l v,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, write l v)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, write l v)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t l v MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  specialize (MapStore l).
  inversion_inverted_lazy_step Step t' l' s r v' v'' Active RgnFresh Fresh Read
    Empty; subst.

  (* tx_write *)
  esplit. split.
  eapply Strong.step_write.
  apply Heap.strong_active_running.
  rewrite VarMapFacts.find_mapsto_iff in Read.
  rewrite Read in MapStore.
  rewrite VarMapFacts.find_mapsto_iff...
  eapply map_heap_write...

  (* write *)
  esplit. split.
  eapply Strong.step_write.
  apply Heap.strong_active_none.
  rewrite VarMapFacts.find_mapsto_iff in Read.
  rewrite Read in MapStore.
  rewrite VarMapFacts.find_mapsto_iff...
  eapply map_heap_write...
Qed.

Lemma step_spawn : forall H3 H1 H3' t t',
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, spawn _ t')) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, spawn _ t')) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t t' MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t'' l s r v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_spawn...
  intro. discriminate.
  auto.
Qed.

Lemma step_makeP : forall H3 H1 H3' t r,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, makeP _ r)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, makeP _ r)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r' v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_makeP.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_makeU : forall H3 H1 H3' t r,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, makeU _ r)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, makeU _ r)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r' v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_makeU.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_makeRO : forall H3 H1 H3' t r,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, makeRO _ r)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, makeRO _ r)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r' v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_makeRO.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_makeTL : forall H3 H1 H3' t r,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, makeTL _ r)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, makeTL _ r)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r' v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_makeTL.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_beginP : forall H3 H1 H3' t,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, beginP _)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, beginP _)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_beginP.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_endP : forall H3 H1 H3' t,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 (Some (t, endP _)) H3' ->
  exists H1',
    Strong.heap_step H1 (Some (t, endP _)) H1'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto.
  intros H3 H1 H3' t MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t' l s r v v' Active RgnFresh Fresh Read
    Empty; subst.
  esplit. split.
  apply Strong.step_endP.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma step_empty : forall H3 H1 H3',
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 None H3' ->
  map_heap_inverted_to_strong H3' H1.
Proof with eauto.
  intros H3 H1 H3' MapHeap Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_inverted_lazy_step Step t l s r v v' Active RgnFresh Fresh Read
    Empty; subst.

  (* commit *)
  eapply map_heap_commit...

  (* end_commit *)
  eapply map_heap_end_commit...
Qed.

Lemma inverted_lazy_to_strong_step  : forall H3 H1 H3' e,
  map_heap_inverted_to_strong H3 H1 ->
  ILazy.heap_step H3 e H3' ->
  exists H1', exists tr,
    erase_trace tr = erase_trace (e :: nil)
    /\ Strong.heap_step_many H1 H1' tr
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto 7.
  intros H3 H1 H3' e MapHeap Step.
  destruct e as [ [ t a ] | ].
  destruct a.

  forwards [ ? [ ? ? ] ] : step_pure...
  forwards [ ? [ ? ? ] ] : step_newRgn...
  forwards [ ? [ ? ? ] ] : step_alloc...
  forwards [ ? [ ? ? ] ] : step_read...
  forwards [ ? [ ? ? ] ] : step_write...
  forwards [ ? [ ? ? ] ] : step_spawn...
  forwards [ ? [ ? ? ] ] : step_makeP...
  forwards [ ? [ ? ? ] ] : step_makeU...
  forwards [ ? [ ? ? ] ] : step_makeRO...
  forwards [ ? [ ? ? ] ] : step_makeTL...
  forwards [ ? [ ? ? ] ] : step_beginP...
  forwards [ ? [ ? ? ] ] : step_endP...

  exists H1. exists (@nil (effect L.val)).
  split... split.
  apply Strong.heap_step_many_refl.
  eapply step_empty...
Qed.

Lemma inverted_lazy_to_strong_many : forall H3 H1 P H3' P' tr,
  map_heap_inverted_to_strong H3 H1 ->
  ILazySys.system_step_many H3 P H3' P' tr ->
  exists H1', exists tr',
    StrongSys.system_step_many H1 P H1' P' tr'
    /\ erase_trace tr = erase_trace tr'
    /\ map_heap_inverted_to_strong H3' H1'.
Proof with eauto 6.
  intros H3 H1 P H3' P' tr MapHeap StepMany.
  induction StepMany as [ ? Step | ? ? ? ? ? ? ? ? ? ? Step ]...
  specializes IHStepMany (>>> __)...
  destruct IHStepMany as [ ? [ ? [ ? [ EraseEq1 ? ] ] ] ].

  inversion Step; subst.
  forwards [ ? [ ? [ EraseEq2 [ ? ? ] ] ] ]
  : inverted_lazy_to_strong_step (>>> __ __)...
  esplit. esplit. split.
  eapply StrongSysFacts.system_step_many_alt...
  eapply StrongSysFacts.system_step_many_mutual...
  split...
  simpl in *.
  rewrite EraseEq1.
  rewrite app_erase.
  rewrite EraseEq2...

  inversion Step; subst.
  forwards [ ? [ ? [ EraseEq2 [ ? ? ] ] ] ]
  : inverted_lazy_to_strong_step (>>> __ __)...
  esplit. esplit. split.
  eapply StrongSysFacts.system_step_many_alt.
  eauto.
  eapply StrongSysFacts.system_step_many_empty...
  split...
  simpl in *.
  rewrite EraseEq1.
  rewrite app_erase.
  rewrite EraseEq2...
Qed.

Lemma inverted_lazy_to_strong : forall P H3 P' tr,
  ILazySys.system_step_many ILazy.empty_heap P H3 P' tr ->
  exists H1, exists tr',
    map_heap_inverted_to_strong H3 H1
    /\ StrongSys.system_step_many Strong.empty_heap P H1 P' tr'.
Proof with eauto.
  intros P H3 P' tr StepMany.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : inverted_lazy_to_strong_many...
  apply map_heap_empty.
Qed.
