Require Import init_modules.

Import Heap.

Inductive map_heap_inverted_to_strong : inverted_heap -> strong_heap -> Prop :=
| map_heap_inverted_to_strong_def  : forall ts3 IS3 R ts1 S1,
  (match ts3 with
     | tx_none
     | tx_commit _ => ts1 = tx_none
     | tx_running t => ts1 = tx_running t
   end) ->
  (forall l,
    match VarMap.find l IS3 with
      | None => VarMap.find l S1 = None
      | Some (_, r, v) => VarMap.find l S1 = Some (r, v)
    end) ->
  map_heap_inverted_to_strong (ts3, IS3, R) (ts1, S1, R).

Lemma map_heap_empty :
  map_heap_inverted_to_strong empty_inverted_heap empty_strong_heap.
Proof with eauto.
  apply map_heap_inverted_to_strong_def...
  intros l.
  rewrite VarMapFacts.empty_o.
  rewrite VarMapFacts.empty_o...
Qed.

Ltac inversion_map_heap MapHeap MapTxState MapStore :=
  inversion MapHeap as [ ? ? ? ? ? MapTxState MapStore ]; subst.

Lemma map_heap_fresh : forall ts3 IS3 R3 ts1 S1 R1 l,
  map_heap_inverted_to_strong (ts3, IS3, R3) (ts1, S1, R1) ->
  (~ VarMap.In l S1 <-> ~ VarMap.In l IS3).
Proof.
  intros ts3 IS3 R3 ts1 S1 R1 l MapHeap.
  inversion_map_heap MapHeap MapTxState MapStore.
  specializes MapStore l.
  split.

  intros Fresh1 [ [ [ s r ] v ] Read3 ].
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  rewrite VarMapFacts.not_find_in_iff in Fresh1.
  rewrite Read3 in MapStore.
  rewrite MapStore in Fresh1.
  inversion Fresh1.

  intros Fresh3.
  rewrite VarMapFacts.not_find_in_iff in Fresh3.
  rewrite VarMapFacts.not_find_in_iff.
  rewrite Fresh3 in MapStore.
  auto.
Qed.

Lemma map_heap_alloc : forall ts3 IS3 R3 ts1 S1 R1 l r v,
  map_heap_inverted_to_strong (ts3, IS3, R3) (ts1, S1, R1) ->
  ~ VarMap.In l IS3 ->
  map_heap_inverted_to_strong (ts3, VarMap.add l (None, r, v) IS3, R3)
  (ts1, VarMap.add l (r, v) S1, R1).
Proof with eauto.
  intros ts3 IS3 R3 ts1 S1 R1 l r v MapHeap Fresh3.
  inversion_map_heap MapHeap MapTxState MapStore.
  apply map_heap_inverted_to_strong_def...
  intros l'.
  specializes MapStore l'.
  destruct (l == l').
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.add_neq_o...
Qed.

Lemma map_heap_tx_read1 : forall ts3 IS3 R3 ts1 S1 R1 l r v,
  map_heap_inverted_to_strong (ts3, IS3, R3) (ts1, S1, R1) ->
  VarMap.MapsTo l (None, r, v) IS3 ->
  map_heap_inverted_to_strong (ts3, VarMap.add l (Some tx_ro, r, v) IS3, R3)
  (ts1, S1, R1).
Proof with auto.
  intros ts3 IS3 R3 ts1 S1 R1 l r v MapHeap Read3.
  inversion_map_heap MapHeap MapTxState MapStore.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  apply map_heap_inverted_to_strong_def...
  intros l'.
  specializes MapStore l'.
  destruct (l == l'). subst.
  rewrite VarMapFacts.add_eq_o...
  rewrite Read3 in MapStore...
  rewrite VarMapFacts.add_neq_o...
Qed.

Lemma map_heap_write : forall ts3 IS3 R3 ts1 S1 R1 l s r v s' v',
  map_heap_inverted_to_strong (ts3, IS3, R3) (ts1, S1, R1) ->
  VarMap.MapsTo l (s, r, v) IS3 ->
  map_heap_inverted_to_strong (ts3, VarMap.add l (s', r, v') IS3, R3)
  (ts1, VarMap.add l (r, v') S1, R1).
Proof with auto.
  intros ts3 IS3 R3 ts1 S1 R1 l s r v s' v' MapHeap Read3.
  inversion_map_heap MapHeap MapTxState MapStore.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  apply map_heap_inverted_to_strong_def...
  intros l'.
  specializes MapStore l'.
  destruct (l == l'). subst.
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.add_neq_o...
Qed.

Lemma map_heap_endP : forall t IS3 R3 S1 R1,
  map_heap_inverted_to_strong (tx_running t, IS3, R3) (tx_running t, S1, R1) ->
  map_heap_inverted_to_strong (tx_commit t, IS3, R3) (tx_none, S1, R1).
Proof with auto.
  intros t IS3 R3 S1 R1 MapHeap.
  inversion_map_heap MapHeap MapTxState MapStore.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma map_heap_commit : forall ts3 IS3 R3 ts1 S1 R1 l s r v,
  map_heap_inverted_to_strong (ts3, IS3, R3) (ts1, S1, R1) ->
  VarMap.MapsTo l (Some s, r, v) IS3 ->
  map_heap_inverted_to_strong (ts3, VarMap.add l (None, r, v) IS3, R3)
  (ts1, S1, R1).
Proof with auto.
  intros ts3 IS3 R3 ts1 S1 R1 l s r v MapHeap Read3.
  inversion_map_heap MapHeap MapTxState MapStore.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  apply map_heap_inverted_to_strong_def...
  intros l'.
  specializes MapStore l'.
  destruct (l == l'). subst.
  rewrite Read3 in MapStore.
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_neq_o...
Qed.

Lemma map_heap_end_commit : forall t IS3 R3 S1 R1,
  map_heap_inverted_to_strong (tx_commit t, IS3, R3) (tx_none, S1, R1) ->
  map_heap_inverted_to_strong (tx_none, IS3, R3) (tx_none, S1, R1).
Proof with auto.
  intros t IS3 R3 S1 R1 MapHeap.
  inversion_map_heap MapHeap MapTxState MapStore.
  apply map_heap_inverted_to_strong_def...
Qed.

Lemma map_heap_read : forall ts3 IS3 R3 ts1 S1 R1 l s r v,
  map_heap_inverted_to_strong (ts3, IS3, R3) (ts1, S1, R1) ->
  VarMap.MapsTo l (s, r, v) IS3 ->
  VarMap.MapsTo l (r, v) S1.
Proof with eauto.
  intros ts3 IS3 R3 ts1 S1 R1 l s r v MapHeap Read.
  inversion_map_heap MapHeap MapTxState MapStore.
  specializes MapStore l.
  rewrite VarMapFacts.find_mapsto_iff in Read.
  rewrite Read in MapStore.
  rewrite VarMapFacts.find_mapsto_iff...
Qed.
