Require Import init_modules.

Import Heap.

Inductive map_heap_inverted_to_weak : inverted_heap -> weak_heap -> Prop :=
| map_heap_inverted_to_weak_def : forall ts IS3 R S2 TS2,
  (forall l,
    match VarMap.find l IS3 with
      | None => VarMap.find l S2 = None /\ VarMap.find l TS2 = None
      | Some (None, r, v) => VarMap.find l S2 = Some (r, v)
        /\ VarMap.find l TS2 = None
      | Some (Some tx_ro, r, v) => VarMap.find l S2 = Some (r, v)
        /\ VarMap.find l TS2 = Some (tx_ro, v)
      | Some (Some tx_mod, r, v) =>
        match VarMap.find l S2 with
          | Some (r', v') => r = r'
          | None => False
        end
        /\ VarMap.find l TS2 = Some (tx_mod, v)
    end) ->
  map_heap_inverted_to_weak (ts, IS3, R) (ts, S2, R, TS2).

Ltac inversion_map_heap MapHeap MapStore :=
  inversion MapHeap as [ ? ? ? ? ? MapStore ]; subst.

Lemma map_heap_inverted_to_weak_empty :
  map_heap_inverted_to_weak empty_inverted_heap empty_weak_heap.
Proof.
  apply map_heap_inverted_to_weak_def.
  intros l.
  repeat rewrite VarMapFacts.empty_o.
  auto.
Qed.

Lemma map_heap_fresh : forall ts3 IS3 R3 ts2 S2 R2 TS2 l,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  (~ VarMap.In l S2 <-> ~ VarMap.In l IS3).
Proof.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l MapHeap.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  split.

  intros Fresh2 [ [ [ s r ] v ] Read3 ].
  subst.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  rewrite VarMapFacts.not_find_in_iff in Fresh2.
  rewrite Read3 in MapStore.
  destruct s as [ [ | ] | ].
  destruct MapStore as [ Read2 ].
  rewrite Fresh2 in Read2.
  inversion Read2.
  rewrite Fresh2 in MapStore.
  destruct MapStore as [ Read2 ].
  inversion Read2.
  destruct MapStore as [ Read2 ].
  rewrite Fresh2 in Read2.
  inversion Read2.

  intros Fresh3.
  rewrite VarMapFacts.not_find_in_iff in Fresh3.
  rewrite VarMapFacts.not_find_in_iff.
  rewrite Fresh3 in MapStore.
  destruct MapStore.
  auto.
Qed.

Lemma map_heap_alloc : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  ~ VarMap.In l IS3 ->
  map_heap_inverted_to_weak (ts3, VarMap.add l (None, r, v) IS3, R3)
  (ts2, VarMap.add l (r, v) S2, R2, TS2).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap Fresh3.
  inversion_map_heap MapHeap MapStore.
  apply map_heap_inverted_to_weak_def.
  intros l'.
  specializes MapStore l'.
  destruct (l == l').
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_eq_o...
  subst l'.
  rewrite VarMapFacts.not_find_in_iff in Fresh3.
  rewrite Fresh3 in MapStore.
  destruct MapStore...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.add_neq_o...
Qed.

Lemma map_heap_read_tx_fresh : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  (VarMap.MapsTo l (r, v) S2 /\ ~ VarMap.In l TS2
    <-> VarMap.MapsTo l (None, r, v) IS3).
Proof.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap.
  rewrite VarMapFacts.find_mapsto_iff.
  rewrite VarMapFacts.not_find_in_iff.
  rewrite VarMapFacts.find_mapsto_iff.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  split.

  intros [ Read2 TxFresh2 ].
  rewrite Read2 in MapStore.
  rewrite TxFresh2 in MapStore.
  destruct (VarMap.find l IS3) as [ [ [ s r' ] v' ] | ].
  destruct s as [ [ | ] | ].
  destruct MapStore as [ ? TxRead ].
  inversion TxRead.
  destruct MapStore as [ ? TxRead ].
  inversion TxRead.
  destruct MapStore as [ Read ].
  inversion Read.
  subst.
  auto.
  destruct MapStore as [ Read ].
  inversion Read.

  intros Read3.
  rewrite Read3 in MapStore.
  auto.
Qed.

Lemma map_heap_tx_read1 : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (None, r, v) IS3 ->
  map_heap_inverted_to_weak (ts3, VarMap.add l (Some tx_ro, r, v) IS3, R3)
  (ts2, S2, R2, VarMap.add l (tx_ro, v) TS2).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap Read3.
  inversion_map_heap MapHeap MapStore.
  apply map_heap_inverted_to_weak_def.
  intros l'.
  specializes MapStore l'.
  destruct (l == l').
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_eq_o...
  subst l'.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  rewrite Read3 in MapStore.
  destruct MapStore...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.add_neq_o...
Qed.

Lemma map_heap_read_tx_read : forall ts3 IS3 R3 ts2 S2 R2 TS2 l s r v v',
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (r, v') S2 ->
  VarMap.MapsTo l (s, v) TS2 ->
  VarMap.MapsTo l (Some s, r, v) IS3.
Proof.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l s r v v' MapHeap Read2 TxRead2.
  rewrite VarMapFacts.find_mapsto_iff in Read2.
  rewrite VarMapFacts.find_mapsto_iff in TxRead2.
  rewrite VarMapFacts.find_mapsto_iff.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  rewrite Read2 in MapStore.
  rewrite TxRead2 in MapStore.
  destruct (VarMap.find l IS3) as [ [ [ s' r' ] v'' ] | ].
  destruct s' as [ [ | ] | ].
  destruct MapStore as [ Read TxRead ].
  inversion Read.
  inversion TxRead.
  subst.
  auto.
  destruct MapStore as [ ? TxRead ].
  inversion TxRead.
  subst.
  auto.
  destruct MapStore as [ ? TxRead ].
  inversion TxRead.
  destruct MapStore as [ Read ].
  inversion Read.
Qed.

Lemma map_heap_tx_read : forall ts3 IS3 R3 ts2 S2 R2 TS2 l s r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (Some s, r, v) IS3 ->
  VarMap.MapsTo l (s, v) TS2 /\ exists v', VarMap.MapsTo l (r, v') S2.
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l s r v MapHeap Read3.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  rewrite VarMapFacts.find_mapsto_iff.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  rewrite Read3 in MapStore.
  destruct s; destruct MapStore as [ Read2 TxRead2 ].
  split...
  esplit. rewrite VarMapFacts.find_mapsto_iff...
  split...
  remember (VarMap.find l S2) as opt.
  destruct opt as [ [ r' v' ] | ].
  subst. esplit. rewrite VarMapFacts.find_mapsto_iff...
  inversion Read2.
Qed.

Lemma map_heap_write : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v v',
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (None, r, v) IS3 ->
  map_heap_inverted_to_weak (ts3, VarMap.add l (None, r, v') IS3, R2)
  (ts2, VarMap.add l (r, v') S2, R2, TS2).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v v' MapHeap Read3.
  inversion_map_heap MapHeap MapStore.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  apply map_heap_inverted_to_weak_def.
  intros l'.
  destruct (l == l').
  repeat rewrite VarMapFacts.add_eq_o...
  subst.
  specializes MapStore l'.
  rewrite Read3 in MapStore.
  destruct MapStore...
  repeat rewrite VarMapFacts.add_neq_o...
  apply MapStore.
Qed.

Lemma map_heap_tx_write : forall ts3 IS3 R3 ts2 S2 R2 TS2 l s r v v',
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (s, r, v) IS3 ->
  map_heap_inverted_to_weak (ts3, VarMap.add l (Some tx_mod, r, v') IS3, R2)
  (ts2, S2, R2, VarMap.add l (tx_mod, v') TS2).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l s r v v' MapHeap Read3.
  inversion_map_heap MapHeap MapStore.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  apply map_heap_inverted_to_weak_def.
  intros l'.
  destruct (l == l').
  repeat rewrite VarMapFacts.add_eq_o...
  subst.
  specializes MapStore l'.
  rewrite Read3 in MapStore.
  destruct s as [ [ | ] | ];
    destruct MapStore as [ Read2 ? ].
  rewrite Read2...
  split...
  rewrite Read2...
  repeat rewrite VarMapFacts.add_neq_o...
  apply MapStore.
Qed.

Lemma map_heap_read : forall ts3 IS3 R3 ts2 S2 R2 TS2 l,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.In l S2 ->
  VarMap.In l IS3.
Proof.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l MapHeap [ [ r v ] Read2 ].
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  rewrite VarMapFacts.find_mapsto_iff in Read2.
  rewrite Read2 in MapStore.
  remember (VarMap.find l IS3) as opt.
  symmetry in Heqopt.
  destruct opt as [ [ [ s r' ] v' ] | ];
    [ rewrite <- VarMapFacts.find_mapsto_iff in Heqopt;
      destruct s as [ [ | ] | ] | ];
      destruct MapStore as [ Eq ];
        inversion Eq; subst; eauto.
Qed.

Lemma map_heap_read_tx_ro : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v v',
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (r, v') S2 ->
  VarMap.MapsTo l (tx_ro, v) TS2 ->
  v = v' /\ VarMap.MapsTo l (Some tx_ro, r, v) IS3.
Proof.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v v' MapHeap Read2 TxRead2.
  rewrite VarMapFacts.find_mapsto_iff in Read2.
  rewrite VarMapFacts.find_mapsto_iff in TxRead2.
  rewrite VarMapFacts.find_mapsto_iff.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  rewrite Read2 in MapStore.
  rewrite TxRead2 in MapStore.
  destruct (VarMap.find l IS3) as [ [ [ s r' ] v'' ] | ].
  destruct s as [ [ | ] | ].
  destruct MapStore as [ Read TxRead ].
  inversion Read.
  inversion TxRead.
  subst.
  auto.
  destruct MapStore as [ ? TxRead ].
  inversion TxRead.
  destruct MapStore as [ ? TxRead ].
  inversion TxRead.
  destruct MapStore as [ Read ].
  inversion Read.
Qed.

Lemma map_heap_tx_ro : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (Some tx_ro, r, v) IS3 ->
  VarMap.MapsTo l (r, v) S2 /\ VarMap.MapsTo l (tx_ro, v) TS2.
Proof.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap Read3.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  rewrite VarMapFacts.find_mapsto_iff.
  rewrite VarMapFacts.find_mapsto_iff.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  rewrite Read3 in MapStore.
  auto.
Qed.

Lemma map_heap_commit_ro : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (Some tx_ro, r, v) IS3 ->
  map_heap_inverted_to_weak (ts3, VarMap.add l (None, r, v) IS3, R3)
  (ts2, S2, R2, VarMap.remove l TS2).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap Read3.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  inversion_map_heap MapHeap MapStore.
  apply map_heap_inverted_to_weak_def.
  intros l'.
  destruct (l == l').
  specializes MapStore l.
  rewrite Read3 in MapStore.
  destruct MapStore as [ Read2 TxRead2 ].
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.remove_eq_o...
  subst...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.remove_neq_o...
  apply MapStore.
Qed.

Lemma map_heap_tx_mod : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (Some tx_mod, r, v) IS3 ->
  VarMap.MapsTo l (tx_mod, v) TS2 /\ exists v', VarMap.MapsTo l (r, v') S2.
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap Read3.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  rewrite VarMapFacts.find_mapsto_iff.
  inversion_map_heap MapHeap MapStore.
  specializes MapStore l.
  rewrite Read3 in MapStore.
  destruct MapStore as [ Read TxRead ].
  split...
  remember (VarMap.find l S2) as opt.
  destruct opt as [ [ r' v' ] | ].
  subst.
  esplit.
  rewrite VarMapFacts.find_mapsto_iff...
  inversion Read.
Qed.

Lemma map_heap_commit_mod : forall ts3 IS3 R3 ts2 S2 R2 TS2 l r v,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  VarMap.MapsTo l (Some tx_mod, r, v) IS3 ->
  map_heap_inverted_to_weak (ts3, VarMap.add l (None, r, v) IS3, R3)
  (ts2, VarMap.add l (r, v) S2, R2, VarMap.remove l TS2).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 l r v MapHeap Read3.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  inversion_map_heap MapHeap MapStore.
  apply map_heap_inverted_to_weak_def.
  intros l'.
  destruct (l == l').
  specializes MapStore l.
  rewrite Read3 in MapStore.
  destruct MapStore as [ Read2 TxRead2 ].
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.remove_eq_o...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.remove_neq_o...
  apply MapStore.
Qed.

Lemma map_heap_tx_empty : forall ts3 IS3 R3 ts2 S2 R2 TS2,
  map_heap_inverted_to_weak (ts3, IS3, R3) (ts2, S2, R2, TS2) ->
  ((forall l, ~ VarMap.In l TS2) <->
    (forall l s r v, VarMap.MapsTo l (s, r, v) IS3 -> s = None)).
Proof with eauto.
  intros ts3 IS3 R3 ts2 S2 R2 TS2 MapHeap.
  inversion_map_heap MapHeap MapStore.
  split.

  intros Empty2 l s r v Read3.
  rewrite VarMapFacts.find_mapsto_iff in Read3.
  specializes MapStore l.
  specializes Empty2 l.
  rewrite Read3 in MapStore.
  destruct s as [ [ | ] | ]...
  destruct MapStore as [ ? TxRead2 ].
  rewrite <- VarMapFacts.find_mapsto_iff in TxRead2.
  contradiction Empty2...
  destruct MapStore as [ ? TxRead2 ].
  rewrite <- VarMapFacts.find_mapsto_iff in TxRead2.
  contradiction Empty2...

  intros Empty3 l [ [ s v ] TxRead2 ].
  rewrite VarMapFacts.find_mapsto_iff in TxRead2.
  specializes MapStore l.
  rewrite TxRead2 in MapStore.
  remember (VarMap.find l IS3) as opt.
  symmetry in Heqopt.
  destruct opt as [ [ [ s' r ] v' ] | ].
  rewrite <- VarMapFacts.find_mapsto_iff in Heqopt.
  specializes Empty3 (>>> Heqopt).
  subst.
  destruct MapStore as [ ? false ].
  inversion false.
  destruct MapStore as [ ? false ].
  inversion false.
Qed.
