Require Import heaps.
Require Import maps.
Require Import models.

Set Implicit Arguments.

Module InvertedWeak (L : LANGUAGE_MODEL) (Heap : HEAP L) <: MEMORY_MODEL L.
  Import Heap.

  Definition heap := inverted_heap.
  Definition empty_heap := empty_inverted_heap.

  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_pure : forall ts IS R t,
    weak_active ts t ->
    step (ts, IS, R) (Some (t, pure _)) (ts, IS, R)
  | step_newRgn : forall ts IS R t r,
    weak_active ts t ->
    ~ VarMap.In r R ->
    step (ts, IS, R) (Some (t, newRgn _ r)) (ts, IS, VarMap.add r pr R)
  | step_alloc : forall ts IS R t l r v,
    weak_active ts t ->
    ~ VarMap.In l IS ->
    step (ts, IS, R) (Some (t, alloc l r v))
    (ts, VarMap.add l (None, r, v) IS, R)
  | step_tx_read1 : forall IS R t l r v,
    VarMap.MapsTo l (None, r, v) IS ->
    step (tx_running t, IS, R) (Some (t, read l v))
    (tx_running t, VarMap.add l (Some tx_ro, r, v) IS, R)
  | step_tx_read2 : forall IS R t l s r v,
    VarMap.MapsTo l (Some s, r, v) IS ->
    step (tx_running t, IS, R) (Some (t, read l v)) (tx_running t, IS, R)
  | step_read : forall IS R t l s r v,
    VarMap.MapsTo l (s, r, v) IS ->
    step (tx_none, IS, R) (Some (t, read l v)) (tx_none, IS, R)
  | step_tx_write : forall IS R t l s r v v',
    VarMap.MapsTo l (s, r, v) IS ->
    step (tx_running t, IS, R) (Some (t, write l v'))
    (tx_running t, VarMap.add l (Some tx_mod, r, v') IS, R)
  | step_write : forall IS R t l s r v v',
    VarMap.MapsTo l (s, r, v) IS ->
    step (tx_none, IS, R) (Some (t, write l v'))
    (tx_none, VarMap.add l (s, r, v') IS, R)
  | step_spawn : forall ts IS R t t',
    weak_unprotected ts t ->
    step (ts, IS, R) (Some (t, spawn _ t')) (ts, IS, R)
  | step_makeP : forall IS R t r,
    step (tx_none, IS, R) (Some (t, makeP _ r)) (tx_none, IS, VarMap.add r pr R)
  | step_makeU : forall IS R t r,
    step (tx_none, IS, R) (Some (t, makeU _ r))
    (tx_none, IS, VarMap.add r unpr R)
  | step_makeRO : forall IS R t r,
    step (tx_none, IS, R) (Some (t, makeRO _ r))
    (tx_none, IS, VarMap.add r ro R)
  | step_makeTL : forall IS R t r,
    step (tx_none, IS, R) (Some (t, makeTL _ r))
    (tx_none, IS, VarMap.add r (tl t) R)
  | step_beginP : forall IS R t,
    step (tx_none, IS, R) (Some (t, beginP _)) (tx_running t, IS, R)
  | step_endP : forall IS R t,
    step (tx_running t, IS, R) (Some (t, endP _)) (tx_commit t, IS, R)
  | step_commit : forall IS R t l s r v,
    VarMap.MapsTo l (Some s, r, v) IS ->
    step (tx_commit t, IS, R) None
    (tx_commit t, VarMap.add l (None, r, v) IS, R)
  | step_end_commit : forall IS R t,
    (forall l s r v,
      VarMap.MapsTo l (s, r, v) IS ->
      s = None) ->
    step (tx_commit t, IS, R) None (tx_none, IS, R).

  Definition heap_step := step.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).
End InvertedWeak.

Ltac inversion_inverted_weak_step Step t l s r v v' Active RgnFresh Fresh Read
  Unprotected Empty :=
  inversion Step as [
    ? ? ? ? Active (* pure *)
    | ? ? ? ? r Active RgnFresh (* newRgn *)
    | ? ? ? ? l r v Active Fresh (* alloc *)
    | ? ? ? l r v Read (* tx_read1 *)
    | ? ? ? l s r v Read (* tx_read2 *)
    | ? ? ? l s r v Read (* read *)
    | ? ? ? l s r v v' Read (* tx_write *)
    | ? ? ? l s r v v' Read (* write *)
    | ? ? ? ? t Unprotected (* spawn *)
    | ? ? ? r (* makeP *)
    | ? ? ? r (* makeU *)
    | ? ? ? r (* makeRO *)
    | ? ? ? r (* makeTL *)
    | (* beginP *)
    | (* endP *)
    | ? ? ? l s r v Read (* commit_ro *)
    | ? ? ? Empty (* end_commit *)
  ]; subst.
