Require Import init_modules.
Require Import inverted_to_weak_map_heap.
Require Import typed_inverted_weak.

Inductive inverted_heap_invariant : IWeak.heap -> Prop :=
| inverted_heap_invariant_def : forall ts3 IS3 R3,
  (match ts3 with
     | Heap.tx_none => forall l s r v,
       VarMap.MapsTo l (s, r, v) IS3 ->
       s = None
     | _ => True
   end) ->
  inverted_heap_invariant (ts3, IS3, R3).

Lemma inverted_heap_invariant_empty :
  inverted_heap_invariant Heap.empty_inverted_heap.
Proof.
  apply inverted_heap_invariant_def.
  intros ? ? ? ? Read.
  rewrite VarMapFacts.empty_mapsto_iff in Read.
  tauto.
Qed.

Lemma inverted_heap_invariant_preservation : forall H3 e H3',
  inverted_heap_invariant H3 ->
  IWeak.heap_step H3 e H3' ->
  inverted_heap_invariant H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] e [ [ ts3' IS3' ] R3' ] Invariant Step.
  inversion Invariant as [ ? ? ? NoLogEntries ]; subst.
  destruct e as [ | [ t a ] ];
    inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty;
    apply inverted_heap_invariant_def...

  (* alloc *)
  destruct ts3'...
  intros l ? ? ? Read.
  rewrite VarMapFacts.add_mapsto_iff in Read.
  destruct Read as [ [ ? Eq ] | [ ? ? ] ]...
  inversion Eq...

  (* write *)
  intros l ? ? ? Read'.
  rewrite VarMapFacts.add_mapsto_iff in Read'.
  destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
  inversion Eq; subst...
Qed.

Lemma step_pure : forall H2 H3 H3' t,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, pure _)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, pure _)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split...
  apply Weak.step_pure...
Qed.

Lemma step_newRgn : forall H2 H3 H3' t r,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, newRgn _ r)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, newRgn _ r)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_newRgn...
  eapply map_heap_inverted_to_weak_def...
Qed.

Lemma step_alloc : forall H2 H3 H3' t l r v,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, alloc l r v)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, alloc l r v)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t l r v MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_alloc...
  eapply map_heap_fresh...
  apply map_heap_alloc...
Qed.

Lemma step_read : forall PT ST H2 H3 H3' t l v,
  TypedIWeak.typing_heap (PT, ST) H3 ->
  inverted_heap_invariant H3 ->
  VarMap.In l ST ->
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, read l v)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, read l v)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros PT ST [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t l v TypingH3
    Invariant TypingLoc MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s r v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.

  esplit. split.
  eapply Weak.step_tx_read1...
  eapply map_heap_read_tx_fresh...
  eapply map_heap_read_tx_fresh...
  apply map_heap_tx_read1...

  inversion_typing_inverted_heap TypingH3 DomStore TypingStore RgnStore DomProt.
  esplit. split.
  eapply Weak.step_tx_read2...
  eapply map_heap_tx_read...
  auto.

  esplit. split...
  eapply Weak.step_read...
  apply Heap.weak_unprotected_none.
  inversion Invariant as [ ? ? ? NoLogEntries ]; subst.
  specializes NoLogEntries (>>> __ __ __ __ __)...
  subst.
  rewrite VarMapFacts.find_mapsto_iff in Read.
  specialize (MapStore l).
  rewrite Read in MapStore.
  destruct MapStore.
  rewrite VarMapFacts.find_mapsto_iff...
Qed.

Lemma step_write : forall PT ST H2 H3 H3' t l v,
  TypedIWeak.typing_heap (PT, ST) H3 ->
  inverted_heap_invariant H3 ->
  VarMap.In l ST ->
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, write l v)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, write l v)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros PT ST [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t l v TypingH3
    Invariant TypingLoc MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.

  inversion_typing_inverted_heap TypingH3 DomStore TypingStore RgnStore DomProt.
  esplit. split.
  eapply Weak.step_tx_write...
  eapply map_heap_tx_write...

  inversion Invariant as [ ? ? ? NoLogEntries ]; subst.
  specializes NoLogEntries (>>> __ __ __ __ __)...
  subst.
  rewrite VarMapFacts.find_mapsto_iff in Read.
  specialize (MapStore l).
  rewrite Read in MapStore.
  destruct MapStore.
  esplit. split.
  eapply Weak.step_write...
  apply Heap.weak_unprotected_none.
  rewrite VarMapFacts.find_mapsto_iff...
  eapply map_heap_write...
  rewrite VarMapFacts.find_mapsto_iff...
Qed.

Lemma step_spawn : forall H2 H3 H3' t t',
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, spawn _ t')) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, spawn _ t')) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t t' MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t'' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_spawn.
  auto. auto.
Qed.

Lemma step_makeP : forall H2 H3 H3' t r,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, makeP _ r)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, makeP _ r)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_makeP.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_makeU : forall H2 H3 H3' t r,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, makeU _ r)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, makeU _ r)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_makeU.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_makeRO : forall H2 H3 H3' t r,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, makeRO _ r)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, makeRO _ r)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_makeRO.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_makeTL : forall H2 H3 H3' t r,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, makeTL _ r)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, makeTL _ r)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_makeTL.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_beginP : forall H2 H3 H3' t,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, beginP _)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, beginP _)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_beginP.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_endP : forall H2 H3 H3' t,
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 (Some (t, endP _)) H3' ->
  exists H2',
    Weak.heap_step H2 (Some (t, endP _)) H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ [ ts2 S2 ] R2 ] TS2 ] [ [ ts3 IS3 ] R3 ] H3' t MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  esplit. split.
  apply Weak.step_endP.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_empty : forall HT H2 H3 H3',
  TypedIWeak.typing_heap HT H3 ->
  map_heap_inverted_to_weak H3 H2 ->
  IWeak.heap_step H3 None H3' ->
  exists H2',
    Weak.heap_step H2 None H2'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros HT H2 H3 H3' TypingH MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_inverted_weak_step Step t' l' s' r' v' v'' Active RgnFresh Fresh
    Read Unprotected Empty.
  destruct s'.

  (* commit_ro *)
  inversion_typing_inverted_heap TypingH DomStore TypingStore RgnStore DomProt.
  esplit. split.
  eapply Weak.step_commit_ro...
  eapply map_heap_tx_ro...
  apply map_heap_commit_ro...

  (* commit_mod *)
  forwards [ ? [ ? ? ] ] : map_heap_tx_mod...
  esplit. split.
  eapply Weak.step_commit_mod...
  apply map_heap_commit_mod...

  (* end_commit *)
  esplit. split.
  apply Weak.step_end_commit.
  eapply map_heap_tx_empty...
  apply map_heap_inverted_to_weak_def...
Qed.

Lemma inverted_weak_to_weak_step : forall HT H2 H3 P e H3' P',
  TypedIWeak.typing_heap HT H3 ->
  Typ.typing_prgm HT P ->
  inverted_heap_invariant H3 ->
  map_heap_inverted_to_weak H3 H2 ->
  IWeakSys.system_step H3 P e H3' P' ->
  exists HT', exists H2',
    TypedIWeak.typing_heap HT' H3'
    /\ Typ.typing_prgm HT' P'
    /\ inverted_heap_invariant H3'
    /\ map_heap_inverted_to_weak H3' H2'
    /\ WeakSys.system_step H2 P e H2' P'.
Proof with eauto 7.
  intros HT H2 H3 P e H3' P' TypingH3 TypingP Invariant MapHeap Step.
  destruct HT as [ PT ST ].
  forwards [ HT' [ Extends [ TypingH3' TypingP' ] ] ] :
    IWeakSysPreservation.type_preservation...
  exists HT'.
  inversion Step as [ ? ? ? ? t a ProgramStep IWeakStep | ? ? ? IWeakStep ];
    subst;
      forwards : inverted_heap_invariant_preservation...

  (* mutual step *)
  destruct a.
  forwards [ ? [ ? ? ] ] : step_pure...
  forwards [ ? [ ? ? ] ] : step_newRgn...
  forwards [ ? [ ? ? ] ] : step_alloc...
  forwards [ ? [ ? ? ] ] : step_read...
  eapply LangReqs.well_formed_read...
  forwards [ ? [ ? ? ] ] : step_write...
  eapply LangReqs.well_formed_write...
  forwards [ ? [ ? ? ] ] : step_spawn...
  forwards [ ? [ ? ? ] ] : step_makeP...
  forwards [ ? [ ? ? ] ] : step_makeU...
  forwards [ ? [ ? ? ] ] : step_makeRO...
  forwards [ ? [ ? ? ] ] : step_makeTL...
  forwards [ ? [ ? ? ] ] : step_beginP...
  forwards [ ? [ ? ? ] ] : step_endP...

  (* empty step *)
  forwards [ ? [ ? ? ] ] : step_empty (>>> TypingH3)...
Qed.

Lemma inverted_weak_to_weak_many : forall HT H2 H3 P H3' P' tr,
  TypedIWeak.typing_heap HT H3 ->
  Typ.typing_prgm HT P ->
  inverted_heap_invariant H3 ->
  map_heap_inverted_to_weak H3 H2 ->
  IWeakSys.system_step_many H3 P H3' P' tr ->
  exists HT', exists H2',
    TypedIWeak.typing_heap HT' H3'
    /\ Typ.typing_prgm HT' P'
    /\ inverted_heap_invariant H3'
    /\ map_heap_inverted_to_weak H3' H2'
    /\ WeakSys.system_step_many H2 P H2' P' tr.
Proof with eauto 8.
  intros HT H2 H3 P H3' P' tr TypingH3 TypingP Invariant MapHeap StepMany.
  generalize dependent HT.
  induction StepMany; intros HT TypingH3 TypingP2; subst...

  forwards [ ? [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] ] : IHStepMany...
  forwards [ ? [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] ] : inverted_weak_to_weak_step...
Qed.

Lemma inverted_weak_to_weak : forall P0 H3 P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  IWeakSys.system_step_many IWeak.empty_heap P0 H3 P tr ->
  exists H2,
    WeakSys.system_step_many Weak.empty_heap P0 H2 P tr.
Proof with eauto.
  intros P0 H3 P tr TypingP0 StepMany.
  forwards TypingH0 : TypedIWeak.typing_empty_heap.
  forwards Invariant0 : inverted_heap_invariant_empty.
  forwards MapHeap0 : map_heap_inverted_to_weak_empty.
  forwards [ ? [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] ] : inverted_weak_to_weak_many...
Qed.
