Require Import maps.
Require Import models.
Require Import types.

Module Type LANGUAGE_REQUIREMENTS.
  Declare Module Typ : TYPE_MODEL with Module HeapTyp := HeapTyping.
  Import Typ.
  Import L.

  Inductive program_order : effect L.val -> effect L.val -> Prop :=
  | program_order_same : forall t a a',
    program_order (Some (t, a)) (Some (t, a'))
  | program_order_spawn : forall t t' a,
    program_order (Some (t, spawn _ t')) (Some (t', a)).

  Axiom program_reordering : forall P1 P2 P3 e e',
    L.program_step P1 e P2 ->
    L.program_step P2 e' P3 ->
    ~ program_order e e' ->
    exists P2', exists P3',
      L.eq_program_state P3 P3'
      /\ L.program_step P1 e' P2'
      /\ L.program_step P2' e P3'.

  Axiom value_independent_reads : forall P t l v P' v',
    L.program_step P (Some (t, read l v)) P' ->
    exists P', L.program_step P (Some (t, read l v')) P'.

  Axiom well_formed_read : forall PT ST P t l v P',
    Typ.typing_prgm (PT, ST) P ->
    L.program_step P (Some (t, read l v)) P' ->
    VarMap.In l ST.

  Axiom well_formed_write : forall PT ST P t l v P',
    Typ.typing_prgm (PT, ST) P ->
    L.program_step P (Some (t, write l v)) P' ->
    VarMap.In l ST.
End LANGUAGE_REQUIREMENTS.
