Require Import init_modules.

Import Heap.
Import Strong.
Import StrongSys.

(** The dynamic separation discipline. *)

Inductive op := op_read | op_write.
Inductive mode := mode_protected | mode_unprotected.

Inductive check : mode -> prot_state -> tid -> op -> Prop :=
| check_unpr : forall t o,
  check mode_unprotected unpr t o
| check_pr : forall t o,
  check mode_protected pr t o
| check_ro : forall m t,
  check m ro t op_read
| check_to : forall m t o,
  check m (tl t) t o.

Inductive separation : L.program_state -> Prop :=
| separation_def : forall P0,
  (forall P tr S R t l r v p H' P',
    system_step_many empty_heap P0 (tx_none, S, R) P tr ->
    system_step (tx_none, S, R) P (Some (t, read l v)) H' P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t op_read) ->
  (forall P tr S R t l r v v' p H' P',
    system_step_many empty_heap P0 (tx_none, S, R) P tr ->
    system_step (tx_none, S, R) P (Some (t, write l v')) H' P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t op_write) ->
  (forall P tr S R t l r v p H' P',
    system_step_many empty_heap P0 (tx_running t, S, R) P tr ->
    system_step (tx_running t, S, R) P (Some (t, read l v)) H' P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_read) ->
  (forall P tr S R t l r v v' p H' P',
    system_step_many empty_heap P0 (tx_running t, S, R) P tr ->
    system_step (tx_running t, S, R) P (Some (t, write l v')) H' P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_write) ->
  separation P0.
