Require Import init_modules.
Require Import inverted_lazy_to_strong.
Require Import inverted_to_strong_map_heap.
Require Import inverted_to_weak_map_heap.
Require Import separation.
Require Import strong_lazy_to_inverted_lazy.

Import Heap.
Module MapStrong := inverted_to_strong_map_heap.
Module MapWeak := inverted_to_weak_map_heap.

Inductive separation_lazy : Lazy.heap -> L.program_state -> Prop :=
| separation_lazy_none : forall S R TS P,
  (forall t l r v p P',
    L.program_step P (Some (t, read l v)) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t op_read) ->
  (forall t l r v v' p P',
    L.program_step P (Some (t, write l v')) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t op_write) ->
  separation_lazy (tx_none, S, R, TS) P
| separation_lazy_running : forall t S R TS P,
  (forall l r v p P',
    L.program_step P (Some (t, read l v)) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    ~ VarMap.In l TS ->
    check mode_protected p t op_read) ->
  (forall l r v s v' p P',
    L.program_step P (Some (t, read l v')) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo l (s, v') TS ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_read) ->
  (forall l r v v' p P',
    L.program_step P (Some (t, write l v')) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_write) ->
  separation_lazy (tx_running t, S, R, TS) P
| separation_lazy_commit : forall t S R TS P,
  separation_lazy (tx_commit t, S, R, TS) P.

Lemma separation_to_separation_lazy : forall P0 H P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  separation P0 ->
  LazySys.system_step_many Lazy.empty_heap P0 H P tr ->
  separation_lazy H P.
Proof with eauto.
  intros P0 [ [ [ ts2 S2 ] R2 ] TS2 ] P tr2 TypingPrgm Separation StepMany2.
  inversion Separation as [ ? InactiveRead InactiveWrite ActiveRead
    ActiveWrite ]; subst.
  forwards TxStoreInvariant : tx_store_invariant_step_many...
  apply tx_store_invariant_empty.
  forwards [ [ [ ts3 IS3 ] R3 ] [ MapHeap32 ? ] ]
  : strong_lazy_to_inverted_lazy...
  MapWeak.inversion_map_heap MapHeap32 MapStore32.
  forwards [ [ [ ts1 S1 ] R1 ] [ ? [ MapHeap31 StepMany1 ] ] ]
  : inverted_lazy_to_strong...
  MapStrong.inversion_map_heap MapHeap31 MapTxState31 MapStore31.
  destruct ts2 as [ | t | t ]; subst ts1.

  (* tx_none *)
  inversion TxStoreInvariant; subst.
  apply separation_lazy_none.

  (* read *)
  intros t l r v p P' PrgmStep Read2 Prot2.
  assert (VarMap.MapsTo l (r, v) S1).
  eapply MapStrong.map_heap_read...
  forwards [ ? ? ] : MapWeak.map_heap_read_tx_fresh...
  eapply InactiveRead...
  apply StrongSys.system_step_mutual...
  eapply Strong.step_read...
  apply strong_active_none.

  (* write *)
  intros t l r v v' p P' PrgmStep Read2 Prot2.
  assert (VarMap.MapsTo l (r, v) S1).
  eapply MapStrong.map_heap_read...
  forwards [ ? ? ] : MapWeak.map_heap_read_tx_fresh...
  eapply InactiveWrite...
  apply StrongSys.system_step_mutual...
  eapply Strong.step_write...
  apply strong_active_none.

  (* tx_running *)
  apply separation_lazy_running.

  (* read1 *)
  intros l r v p P' PrgmStep Read2 Prot2 TxFresh2.
  assert (VarMap.MapsTo l (r, v) S1).
  eapply MapStrong.map_heap_read...
  forwards [ ? ? ] : MapWeak.map_heap_read_tx_fresh...
  eapply ActiveRead...
  apply StrongSys.system_step_mutual...
  eapply Strong.step_read...
  apply strong_active_running.

  (* read2 *)
  intros l r v s v' p P' PrgmStep Read2 TxRead2 Prot2.
  assert (VarMap.MapsTo l (r, v') S1).
  eapply MapStrong.map_heap_read...
  eapply MapWeak.map_heap_read_tx_read...
  eapply ActiveRead...
  apply StrongSys.system_step_mutual...
  eapply Strong.step_read...
  apply strong_active_running.

  (* write1 *)
  intros l r v v' p P' PrgmStep Read2 Prot2.
  assert (exists v'', VarMap.MapsTo l (r, v'') S1) as [ ? ? ].
  pose proof (@VarMapFacts.In_dec (tx_loc_state * L.val)) TS2 l
    as [ [ [ ? ? ] ? ] | ? ].
  esplit. eapply MapStrong.map_heap_read...
  eapply MapWeak.map_heap_read_tx_read...
  esplit. eapply MapStrong.map_heap_read...
  forwards [ ? ? ] : MapWeak.map_heap_read_tx_fresh...
  eapply ActiveWrite...
  apply StrongSys.system_step_mutual...
  eapply Strong.step_write...
  apply strong_active_running.

  (* tx_commit *)
  apply separation_lazy_commit.
Qed.
