Require Import init_modules.
Require Import separation.
Require Import separation_lazy.
Require Import weak_heap_invariant.
Require Import typed_strong_lazy.

Import Heap.

Set Implicit Arguments.

Ltac not_program_order :=
  intro ProgramOrder; inversion ProgramOrder; subst; tauto.

Inductive separation_weak : Weak.heap -> L.program_state -> Prop :=
| separation_weak_none : forall S R TS P,
  (forall t l r v p P',
    L.program_step P (Some (t, read l v)) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t op_read) ->
  (forall t l r v v' p P',
    L.program_step P (Some (t, write l v')) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t op_write) ->
  separation_weak (tx_none, S, R, TS) P
| separation_weak_running : forall t S R TS P,
  (forall t' l r v p P',
    L.program_step P (Some (t', read l v)) P' ->
    t <> t' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t' op_read) ->
  (forall t' l r v v' p P',
    L.program_step P (Some (t', write l v')) P' ->
    t <> t' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t' op_write) ->
  (forall l r v p P',
    L.program_step P (Some (t, read l v)) P' ->
    VarMap.MapsTo l (r, v) S ->
    ~ VarMap.In l TS ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_read) ->
  (forall l s r v v' p P',
    L.program_step P (Some (t, read l v)) P' ->
    VarMap.MapsTo l (s, v) TS ->
    VarMap.MapsTo l (r, v') S ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_read) ->
  (forall l r v v' p P',
    L.program_step P (Some (t, write l v')) P' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_protected p t op_write) ->
  separation_weak (tx_running t, S, R, TS) P
| separation_weak_commit : forall t S R TS P,
  (forall t' l r v p P',
    L.program_step P (Some (t', read l v)) P' ->
    t <> t' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t' op_read) ->
  (forall t' l r v v' p P',
    L.program_step P (Some (t', write l v')) P' ->
    t <> t' ->
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo r p R ->
    check mode_unprotected p t' op_write) ->
  separation_weak (tx_commit t, S, R, TS) P.

Ltac inversion_separation_weak SeparationWeak :=
  inversion SeparationWeak as [ ? ? ? ? ReadInactive WriteInactive
    | ? ? ? ? ? ReadInactive WriteInactive ReadActive1
      ReadActive2 WriteActive2 | ? ? ? ? ? ReadInactive WriteInactive ]; subst.

Ltac inversion_separation_lazy SeparationLazy :=
  inversion SeparationLazy as [ ? ? ? ? ReadInactive' WriteInactive'
    | ? ? ? ? ? ReadActive1' ReadActive2' WriteActive2' | ]; subst.

Lemma separation_weak_empty : forall P0,
  separation_weak Weak.empty_heap P0.
Proof.
  intros P0.
  apply separation_weak_none.
  intros t l r v p P Step Read.
  rewrite VarMapFacts.empty_mapsto_iff in Read.
  tauto.
  intros t l r v v' p P Step Read Prot.
  rewrite VarMapFacts.empty_mapsto_iff in Read.
  tauto.
Qed.

Lemma separation_weak_eq : forall H P H' P',
  separation_weak H P ->
  eq_weak_heap H H' ->
  L.eq_program_state P P' ->
  separation_weak H' P'.
Proof with eauto.
  intros [ [ [ ts S ] R ] TS ] P [ [ [ ts' S' ] R' ] TS' ] P' SeparationWeak
    HeapEq PrgmEq.
  inversion SeparationWeak as
    [ ? ? ? ? ReadInactive WriteInactive
      | ? ? ? ? ? ReadInactive WriteInactive ReadActive ReadActive' WriteActive
      | ? ? ? ? ? ReadInactive WriteInactive
    ]; subst; inversion HeapEq; subst.

  apply separation_weak_none.
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply ReadInactive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply WriteInactive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...

  apply separation_weak_running.
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply ReadInactive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply WriteInactive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply ReadActive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.In_m...
  rewrite VarMapFacts.MapsTo_m...
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply ReadActive'...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply WriteActive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...

  apply separation_weak_commit.
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply ReadInactive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
  intros.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step... symmetry...
  eapply WriteInactive...
  rewrite VarMapFacts.MapsTo_m...
  rewrite VarMapFacts.MapsTo_m...
Qed.

Lemma separation_lazy_to_weak_none : forall S1 R1 TS1 P1,
  separation_lazy (tx_none, S1, R1, TS1) P1 ->
  separation_weak (tx_none, S1, R1, TS1) P1.
Proof with eauto.
  intros S1 R1 TS1 P1 Separation.
  inversion_separation_lazy Separation.
  apply separation_weak_none...
Qed.

Lemma separation_lazy_to_weak_tx_pure : forall t S1 R1 TS1 P1 P2,
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  L.program_step P1 (Some (t, pure _)) P2 ->
  separation_lazy (tx_running t, S1, R1, TS1) P2 ->
  separation_weak (tx_running t, S1, R1, TS1) P2.
Proof with eauto.
  intros t S1 R1 TS1 P1 P2 SeparationWeak1 ProgramStep2 SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_running...

  intros t' l r v p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...

  intros t' l r v v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_tx_newRgn : forall HT t S1 R1 TS1 P1 P2 r,
  TypedStrongLazy.typing_heap HT (tx_running t, S1, R1, TS1) ->
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  ~ VarMap.In r R1 ->
  L.program_step P1 (Some (t, newRgn _ r)) P2 ->
  separation_lazy (tx_running t, S1, VarMap.add r pr R1, TS1) P2 ->
  separation_weak (tx_running t, S1, VarMap.add r pr R1, TS1) P2.
Proof with eauto.
  intros HT t S1 R1 TS1 P1 P2 r TypingHeap SeparationWeak1 RgnFresh1
    ProgramStep2 SeparationLazy2.
  inversion_typing_weak_heap TypingHeap DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_running...

  intros t' l r' v p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...
  rewrite <- VarMapFacts.add_neq_mapsto_iff...
  intro. subst. apply RgnFresh1.
  rewrite <- DomProt.
  eapply RgnStore...

  intros t' l r' v v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
  rewrite <- VarMapFacts.add_neq_mapsto_iff...
  intro. subst. apply RgnFresh1.
  rewrite <- DomProt.
  eapply RgnStore...
Qed.

Lemma separation_lazy_to_weak_tx_alloc : forall HT t S1 R1 TS1 P1 P2 l r v,
  Typ.typing_prgm HT P1 ->
  TypedStrongLazy.typing_heap HT (tx_running t, S1, R1, TS1) ->
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  ~ VarMap.In l S1 ->
  L.program_step P1 (Some (t, alloc l r v)) P2 ->
  separation_lazy (tx_running t, VarMap.add l (r, v) S1, R1, TS1) P2 ->
  separation_weak (tx_running t, VarMap.add l (r, v) S1, R1, TS1) P2.
Proof with eauto.
  intros [ PT ST ] t S1 R1 TS1 P1 P2 l r v TypingPrgm TypingHeap SeparationWeak1
    Fresh1 ProgramStep2 SeparationLazy2.
  inversion_typing_weak_heap TypingHeap DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.

  apply separation_weak_running...

  intros t' l' r' v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...
  forwards : LangReqs.well_formed_read...
  rewrite <- VarMapFacts.add_neq_mapsto_iff...
  intro. subst. apply Fresh1.
  rewrite <- DomStore...

  intros t' l' r' v' v'' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
  forwards : LangReqs.well_formed_write...
  rewrite <- VarMapFacts.add_neq_mapsto_iff...
  intro. subst. apply Fresh1.
  rewrite <- DomStore...
Qed.

Lemma separation_lazy_to_weak_tx_read1 : forall t S1 R1 TS1 P1 P2 l r v,
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  ~ VarMap.In l TS1 ->
  VarMap.MapsTo l (r, v) S1 ->
  L.program_step P1 (Some (t, read l v)) P2 ->
  separation_lazy (tx_running t, S1, R1, VarMap.add l (tx_ro, v) TS1) P2 ->
  separation_weak (tx_running t, S1, R1, VarMap.add l (tx_ro, v) TS1) P2.
Proof with eauto.
  intros t S1 R1 TS1 P1 P2 l r v SeparationWeak1 TxFresh1 Read1 ProgramStep2
    SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_running...

  intros t' l' r' v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...

  intros t' l' r' v' v'' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_tx_read2 : forall t S1 R1 TS1 P1 P2 l s v,
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  VarMap.MapsTo l (s, v) TS1 ->
  L.program_step P1 (Some (t, read l v)) P2 ->
  separation_lazy (tx_running t, S1, R1, TS1) P2 ->
  separation_weak (tx_running t, S1, R1, TS1) P2.
Proof with eauto.
  intros t S1 R1 TS1 P1 P2 l s v SeparationWeak1 Read1 ProgramStep2
    SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_running...

  intros t' l' r' v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...

  intros t' l' r' v' v'' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_tx_write : forall t S1 R1 TS1 P1 P2 l v,
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  L.program_step P1 (Some (t, write l v)) P2 ->
  separation_lazy (tx_running t, S1, R1, VarMap.add l (tx_mod, v) TS1) P2 ->
  separation_weak (tx_running t, S1, R1, VarMap.add l (tx_mod, v) TS1) P2.
Proof with eauto.
  intros t S1 R1 TS1 P1 P2 l v SeparationWeak1 ProgramStep2 SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_running...

  intros t' l' r' v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...

  intros t' l' r' v' v'' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_tx_beginP : forall S1 R1 TS1 P1 t P2,
  separation_weak (tx_none, S1, R1, TS1) P1 ->
  L.program_step P1 (Some (t, beginP _)) P2 ->
  separation_lazy (tx_running t, S1, R1, TS1) P2 ->
  separation_weak (tx_running t, S1, R1, TS1) P2.
Proof with eauto.
  intros S1 R1 TS1 P1 t P2 SeparationWeak1 ProgramStep2 SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_running...

  intros t' l' r' v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...

  intros t' l' r' v' v'' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_running : forall HT H1 P1 e1 t S2 R2 TS2 P2,
  Typ.typing_prgm HT P1 ->
  TypedStrongLazy.typing_heap HT H1 ->
  separation_weak H1 P1 ->
  LazySys.system_step H1 P1 e1 (tx_running t, S2, R2, TS2) P2 ->
  separation_lazy (tx_running t, S2, R2, TS2) P2 ->
  separation_weak (tx_running t, S2, R2, TS2) P2.
Proof with eauto.
  intros HT H1 P1 e1 t S2 R2 TS2 P2 TypingPrgm TypingHeap SeparationWeak1
    LazySysStep2 SeparationLazy2.
  inversion LazySysStep2 as [ ? ? ? ? ? ? ? LazyStep2 | ? ? ? LazyStep2 ];
    inversion_lazy_step LazyStep2 t' l r v v' Active RgnFresh Fresh Read TxFresh
    TxRead Empty; subst.
  inversion Active. subst.
  eapply separation_lazy_to_weak_tx_pure...
  inversion Active. subst.
  eapply separation_lazy_to_weak_tx_newRgn...
  inversion Active. subst.
  eapply separation_lazy_to_weak_tx_alloc...
  eapply separation_lazy_to_weak_tx_read1...
  eapply separation_lazy_to_weak_tx_read2...
  eapply separation_lazy_to_weak_tx_write...
  eapply separation_lazy_to_weak_tx_beginP...
Qed.

Lemma separation_lazy_to_weak_endP : forall t S1 R1 TS1 P1 P2,
  separation_weak (tx_running t, S1, R1, TS1) P1 ->
  L.program_step P1 (Some (t, endP _)) P2 ->
  separation_lazy (tx_commit t, S1, R1, TS1) P2 ->
  separation_weak (tx_commit t, S1, R1, TS1) P2.
Proof with eauto.
  intros t S1 R1 TS1 P1 P2 SeparationWeak1 ProgramStep2 SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion_separation_lazy SeparationLazy2.
  apply separation_weak_commit.

  intros t' l r v p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply ReadInactive...

  intros t' l r v v' p P3 ProgramStep3 Neq Read Prot.
  forwards [ P2' [ P3' [ ? [ ProgramStep2' ProgramStep3' ] ] ] ]
    : LangReqs.program_reordering (>>> ProgramStep2 ProgramStep3).
  not_program_order.
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_commit_ro : forall t S1 R1 TS1 P1 l v,
  separation_weak (tx_commit t, S1, R1, TS1) P1 ->
  VarMap.MapsTo l (tx_ro, v) TS1 ->
  separation_lazy (tx_commit t, S1, R1, VarMap.remove l TS1) P1 ->
  separation_weak (tx_commit t, S1, R1, VarMap.remove l TS1) P1.
Proof with eauto.
  intros t S1 R1 TS1 P1 l v SeparationWeak1 TxRead1 SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  apply separation_weak_commit...
Qed.

Lemma separation_lazy_to_weak_commit_mod : forall HT t S1 R1 TS1 P1 l r v v',
  TypedStrongLazy.typing_heap HT (tx_commit t, S1, R1, TS1) ->
  weak_heap_invariant (tx_commit t, S1, R1, TS1) ->
  separation_weak (tx_commit t, S1, R1, TS1) P1 ->
  VarMap.MapsTo l (r, v) S1 ->
  VarMap.MapsTo l (tx_mod, v') TS1 ->
  separation_lazy (tx_commit t, VarMap.add l (r, v') S1, R1,
    VarMap.remove l TS1) P1 ->
  separation_weak (tx_commit t, VarMap.add l (r, v') S1, R1,
    VarMap.remove l TS1) P1.
Proof with eauto.
  intros [ PT ST ] t S1 R1 TS1 P1 l r v v' TypingHeap Invariant1 SeparationWeak1
    Read1 TxRead1 SeparationLazy2.
  inversion_separation_weak SeparationWeak1.
  inversion Invariant1 as [ ? ? ? ? TxROState TxModState ]; subst.
  inversion_typing_weak_heap TypingHeap DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  forwards Prot1 : RgnStore (>>> Read1).
  rewrite DomProt in Prot1.
  destruct Prot1 as [ p Prot1 ].
  forwards CheckState1 : TxModState (>>> __ __ __)...
  apply separation_weak_commit.

  intros t' l' r' v'' p' P2 ProgramStep2 Neq Read1' Prot1'.
  rewrite VarMapFacts.add_mapsto_iff in Read1'.
  destruct Read1' as [ [ ? Eq ] | [ ? ? ] ].
  inversion Eq; subst.
  forwards : VarMapFacts.MapsTo_fun (>>> Prot1 Prot1'); subst.
  forwards [ P2' ProgramStep2' ]: LangReqs.value_independent_reads...
  eapply ReadInactive...

  intros t' l' r' v'' v''' p' P2 ProgramStep2 Neq Read1' Prot1'.
  rewrite VarMapFacts.add_mapsto_iff in Read1'.
  destruct Read1' as [ [ ? Eq ] | [ ? ? ] ].
  inversion Eq; subst.
  forwards : VarMapFacts.MapsTo_fun (>>> Prot1 Prot1'); subst...
  eapply WriteInactive...
Qed.

Lemma separation_lazy_to_weak_commit : forall HT H1 P1 e1 t S2 R2 TS2 P2,
  TypedStrongLazy.typing_heap HT H1 ->
  weak_heap_invariant H1 ->
  separation_weak H1 P1 ->
  LazySys.system_step H1 P1 e1 (tx_commit t, S2, R2, TS2) P2 ->
  separation_lazy (tx_commit t, S2, R2, TS2) P2 ->
  separation_weak (tx_commit t, S2, R2, TS2) P2.
Proof with eauto.
  intros HT H1 P1 e1 t S2 R2 TS2 P2 TypingHeap1 Invariant1 SeparationWeak1
    LazySysStep2 SeparationLazy2.
  inversion LazySysStep2 as [ ? ? ? ? ? ? ? LazyStep2 | ? ? ? LazyStep2 ];
    inversion_lazy_step LazyStep2 t' l r v v' Active RgnFresh Fresh Read TxFresh
      TxRead Empty;
        try inversion Active.
  eapply separation_lazy_to_weak_endP...
  eapply separation_lazy_to_weak_commit_ro...
  eapply separation_lazy_to_weak_commit_mod...
Qed.

Lemma separation_lazy_to_weak : forall HT H1 P1 e1 H2 P2,
  Typ.typing_prgm HT P1 ->
  TypedStrongLazy.typing_heap HT H1 ->
  weak_heap_invariant H1 ->
  separation_weak H1 P1 ->
  LazySys.system_step H1 P1 e1 H2 P2 ->
  separation_lazy H2 P2 ->
  separation_weak H2 P2.
Proof with eauto.
  intros HT H1 P1 e1 [ [ [ [ | t | t ] S2 ] R2 ] TS2 ] P2 TypingPrgm TypingHeap
    Invariant1 SeparationWeak1 LazySysStep SeparationLazy2.
  eapply separation_lazy_to_weak_none...
  eapply separation_lazy_to_weak_running...
  eapply separation_lazy_to_weak_commit...
Qed.

Lemma separation_lazy_to_weak_many : forall P0 H P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  separation P0 ->
  LazySys.system_step_many empty_weak_heap P0 H P tr ->
  separation_weak H P.
Proof with eauto.
  intros P0 H P tr TypingPrgm0 Separation0 LazySysStepMany.

  forwards SeparationLazy0 : separation_to_separation_lazy.
  eauto. auto. apply LazySys.system_step_many_refl.

  unfold Lazy.empty_heap in *.
  remember empty_weak_heap as H0.
  remember (HeapTyping.empty_heap_typing Typ.typ) as HT.
  generalize dependent HT.
  induction LazySysStepMany; intros HT HeqHT TypingPrgm.

  subst. apply separation_weak_empty.

  specializes IHLazySysStepMany (>>> __ __ __ __ __ __)...
  subst.
  forwards [ ? [ TypingHeap2 TypingPrgm2 ] ] :
    LazySysPreservation.type_preservation_many...
  eapply separation_lazy_to_weak...
  eapply weak_heap_invariant_step_many...
  eapply separation_to_separation_lazy...
  eapply LazySys.system_step_many_trans...
Qed.

