Require Import heaps.
Require Import maps.
Require Import models.

Set Implicit Arguments.

Module Strong (L : LANGUAGE_MODEL) (Heap : HEAP L) <: MEMORY_MODEL L.
  Import Heap.

  Definition heap := strong_heap.
  Definition empty_heap := empty_strong_heap.

  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_pure : forall H t,
    step H (Some (t, pure _)) H
  | step_newRgn : forall ts S R t r,
    ~ VarMap.In r R ->
    step (ts, S, R) (Some (t, newRgn _ r)) (ts, S, VarMap.add r pr R)
  | step_alloc : forall ts S R t l r v,
    ~ VarMap.In l S ->
    step (ts, S, R) (Some (t, alloc l r v)) (ts, VarMap.add l (r, v) S, R)
  | step_read : forall ts S R t l r v,
    strong_active ts t ->
    VarMap.MapsTo l (r, v) S ->
    step (ts, S, R) (Some (t, read l v)) (ts, S, R)
  | step_write : forall ts S R t l r v v',
    strong_active ts t ->
    VarMap.MapsTo l (r, v) S ->
    step (ts, S, R) (Some (t, write l v')) (ts, VarMap.add l (r, v') S, R)
  | step_spawn : forall ts S R t t',
    ts <> tx_running t ->
    step (ts, S, R) (Some (t, spawn _ t')) (ts, S, R)
  | step_makeP : forall S R t r,
    step (tx_none, S, R) (Some (t, makeP _ r))
    (tx_none, S, VarMap.add r pr R)
  | step_makeU : forall S R t r,
    step (tx_none, S, R) (Some (t, makeU _ r))
    (tx_none, S, VarMap.add r unpr R)
  | step_makeRO : forall S R t r,
    step (tx_none, S, R) (Some (t, makeRO _ r))
    (tx_none, S, VarMap.add r ro R)
  | step_makeTL : forall S R t r,
    step (tx_none, S, R) (Some (t, makeTL _ r))
    (tx_none, S, VarMap.add r (tl t) R)
  | step_beginP : forall S R t,
    step (tx_none, S, R) (Some (t, beginP _)) (tx_running t, S, R)
  | step_endP : forall S R t,
    step (tx_running t, S, R) (Some (t, endP _)) (tx_none, S, R).

  Definition heap_step := step.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).
End Strong.

Ltac inversion_strong_step Step t l r v v' RgnFresh Fresh Active Read
  NotRunning :=
  inversion Step as [
    (* pure *)
    | ? ? ? ? r RgnFresh (* newRgn *)
    | ? ? ? ? l r v Fresh (* alloc *)
    | ? ? ? ? l r v Active Read (* read *)
    | ? ? ? ? l r v v' Active Read (* write *)
    | ? ? ? ? t NotRunning (* spawn *)
    | ? ? ? r (* makeP *)
    | ? ? ? r (* makeU *)
    | ? ? ? r (* makeRO *)
    | ? ? ? r (* makeTL *)
    | (* beginP *)
    | (* endP *)
  ]; subst.