Require Import init_modules.
Require Import inverted_to_weak_map_heap.
Require Import typed_strong_lazy.

Inductive tx_store_invariant : Lazy.heap -> Prop :=
| tx_store_invariant_none : forall S R TS,
  (forall l, ~ VarMap.In l TS) ->
  tx_store_invariant (Heap.tx_none, S, R, TS)
| tx_store_invariant_running : forall t S R TS,
  tx_store_invariant (Heap.tx_running t, S, R, TS)
| tx_store_invariant_commit : forall t S R TS,
  tx_store_invariant (Heap.tx_commit t, S, R, TS).

Hint Constructors tx_store_invariant.

Lemma tx_store_invariant_empty :
  tx_store_invariant Lazy.empty_heap.
Proof.
  apply tx_store_invariant_none.
  intros l false.
  rewrite VarMapFacts.empty_in_iff in false.
  auto.
Qed.

Lemma tx_store_invariant_step : forall H2 e H2',
  tx_store_invariant H2 ->
  Lazy.heap_step H2 e H2' ->
  tx_store_invariant H2'.
Proof with eauto.
  intros H2 e H2' Invariant Step.
  inversion_lazy_step Step t l r v v' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst;
      inversion Invariant; subst...
Qed.

Lemma tx_store_invariant_step_many : forall H2 P H2' P' tr,
  tx_store_invariant H2 ->
  LazySys.system_step_many H2 P H2' P' tr ->
  tx_store_invariant H2'.
Proof with eauto.
  intros H2 P H2' P' tr Invariant StepMany.
  induction StepMany as [ | ? ? ? ? ? ? ? ? StepMany ? Step ]...
  inversion Step; subst.
  eapply tx_store_invariant_step.
  apply IHStepMany...
  eauto.
  eapply tx_store_invariant_step.
  apply IHStepMany...
  eauto.
Qed.

Lemma step_pure : forall H3 H2 H2' t,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, pure _)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, pure _)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split...
  apply ILazy.step_pure...
Qed.

Lemma step_newRgn : forall H3 H2 H2' t r,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, newRgn _ r)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, newRgn _ r)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_newRgn...
  eapply map_heap_inverted_to_weak_def...
Qed.

Lemma step_alloc : forall H3 H2 H2' t l r v,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, alloc l r v)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, alloc l r v)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t l r v MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_alloc...
  eapply map_heap_fresh...
  apply map_heap_alloc...
  eapply map_heap_fresh...
Qed.

Lemma step_read : forall PT ST H3 H2 H2' t l v,
  TypedStrongLazy.typing_heap (PT, ST) H2 ->
  VarMap.In l ST ->
  tx_store_invariant H2 ->
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, read l v)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, read l v)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros PT ST [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t l v TypingH2
    TypingLoc Invariant MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t' l' r v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.

  esplit. split.
  apply ILazy.step_tx_read1...
  eapply map_heap_read_tx_fresh...
  apply map_heap_tx_read1...
  eapply map_heap_read_tx_fresh...

  inversion_typing_weak_heap TypingH2 DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  assert (VarMap.In l S2) as [ [ r v' ] Read2 ].
  rewrite <- DomStore...
  esplit. split.
  eapply ILazy.step_tx_read2...
  eapply map_heap_read_tx_read...
  auto.

  esplit. split...
  eapply ILazy.step_read...
  eapply map_heap_read_tx_fresh...
  inversion Invariant...
Qed.

Lemma step_write : forall PT ST H3 H2 H2' t l v,
  TypedStrongLazy.typing_heap (PT, ST) H2 ->
  VarMap.In l ST ->
  tx_store_invariant H2 ->
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, write l v)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, write l v)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros PT ST [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t l v TypingH2
    TypingLoc Invariant MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t' l' r v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.

  inversion_typing_weak_heap TypingH2 DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  forwards [ [ [ ? ? ] ? ] ? ] : map_heap_read...
  rewrite <- DomStore...
  esplit. split.
  eapply ILazy.step_tx_write...
  eapply map_heap_tx_write...

  inversion Invariant; subst.
  forwards [ ? ? ] : map_heap_read_tx_fresh...
  esplit. split.
  eapply ILazy.step_write...
  eapply map_heap_write...
Qed.

Lemma step_spawn : forall H3 H2 H2' t t',
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, spawn _ t')) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, spawn _ t')) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t t' MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_spawn.
  auto.
Qed.

Lemma step_makeP : forall H3 H2 H2' t r,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, makeP _ r)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, makeP _ r)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_makeP.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_makeU : forall H3 H2 H2' t r,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, makeU _ r)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, makeU _ r)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_makeU.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_makeRO : forall H3 H2 H2' t r,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, makeRO _ r)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, makeRO _ r)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_makeRO.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_makeTL : forall H3 H2 H2' t r,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, makeTL _ r)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, makeTL _ r)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t r MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_makeTL.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_beginP : forall H3 H2 H2' t,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, beginP _)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, beginP _)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_beginP.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_endP : forall H3 H2 H2' t,
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 (Some (t, endP _)) H2' ->
  exists H3',
    ILazy.heap_step H3 (Some (t, endP _)) H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof.
  intros [ [ ts3 IS3 ] R3 ] [ [ [ ts2 S2 ] R2 ] TS2 ] H2' t MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t'' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.
  esplit. split.
  apply ILazy.step_endP.
  apply map_heap_inverted_to_weak_def...
  apply MapStore.
Qed.

Lemma step_empty : forall HT H3 H2 H2',
  TypedStrongLazy.typing_heap HT H2 ->
  map_heap_inverted_to_weak H3 H2 ->
  Lazy.heap_step H2 None H2' ->
  exists H3',
    ILazy.heap_step H3 None H3'
    /\ map_heap_inverted_to_weak H3' H2'.
Proof with eauto.
  intros HT H3 H2 H2' TypingH MapHeap Step.
  inversion_map_heap MapHeap MapStore.
  inversion_lazy_step Step t l r v v' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Empty2; subst.

  (* commit_ro *)
  inversion_typing_weak_heap TypingH DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  specializes DomTx (>>> __ __)...
  assert (VarMap.In l S2) as [ [ r v' ] Read2 ].
  rewrite <- DomStore...
  forwards [ ? ? ] : map_heap_read_tx_ro...
  subst v'.
  esplit. split.
  eapply ILazy.step_commit...
  apply map_heap_commit_ro...

  (* commit_mod *)
  forwards : map_heap_read_tx_read...
  esplit. split.
  eapply ILazy.step_commit...
  apply map_heap_commit_mod...

  (* end_commit *)
  esplit. split.
  apply ILazy.step_end_commit.
  eapply map_heap_tx_empty...
  apply map_heap_inverted_to_weak_def...
Qed.

Lemma strong_lazy_to_inverted_lazy_step : forall HT H3 H2 P e H2' P',
  TypedStrongLazy.typing_heap HT H2 ->
  Typ.typing_prgm HT P ->
  tx_store_invariant H2 ->
  map_heap_inverted_to_weak H3 H2 ->
  LazySys.system_step H2 P e H2' P' ->
  exists HT', exists H3',
    TypedStrongLazy.typing_heap HT' H2'
    /\ Typ.typing_prgm HT' P'
    /\ map_heap_inverted_to_weak H3' H2'
    /\ ILazySys.system_step H3 P e H3' P'.
Proof with eauto 7.
  intros HT H3 H2 P e H2' P' TypingH2 TypingP Invariant MapHeap Step.
  destruct HT as [ PT ST ].
  forwards [ HT' [ Extends [ TypingH2' TypingP' ] ] ] :
    LazySysPreservation.type_preservation...
  exists HT'.
  inversion Step as [ ? ? ? ? t a ProgramStep LazyStep | ? ? ? LazyStep ];
    subst.

  (* mutual step *)
  destruct a.
  forwards [ ? [ ? ? ] ] : step_pure...
  forwards [ ? [ ? ? ] ] : step_newRgn...
  forwards [ ? [ ? ? ] ] : step_alloc...
  forwards [ ? [ ? ? ] ] : step_read...
  eapply LangReqs.well_formed_read...
  forwards [ ? [ ? ? ] ] : step_write...
  eapply LangReqs.well_formed_write...
  forwards [ ? [ ? ? ] ] : step_spawn...
  forwards [ ? [ ? ? ] ] : step_makeP...
  forwards [ ? [ ? ? ] ] : step_makeU...
  forwards [ ? [ ? ? ] ] : step_makeRO...
  forwards [ ? [ ? ? ] ] : step_makeTL...
  forwards [ ? [ ? ? ] ] : step_beginP...
  forwards [ ? [ ? ? ] ] : step_endP...

  (* empty step *)
  forwards [ ? [ ? ? ] ] : step_empty (>>> TypingH2)...
Qed.

Lemma strong_lazy_to_inverted_lazy_many : forall HT H3 H2 P H2' P' tr,
  TypedStrongLazy.typing_heap HT H2 ->
  Typ.typing_prgm HT P ->
  tx_store_invariant H2 ->
  map_heap_inverted_to_weak H3 H2 ->
  LazySys.system_step_many H2 P H2' P' tr ->
  exists HT', exists H3',
    TypedStrongLazy.typing_heap HT' H2'
    /\ Typ.typing_prgm HT' P'
    /\ map_heap_inverted_to_weak H3' H2'
    /\ ILazySys.system_step_many H3 P H3' P' tr.
Proof with eauto 8.
  intros HT H3 H2 P H2' P' tr TypingH2 TypingP Invariant MapHeap StepMany.
  generalize dependent HT.
  induction StepMany; intros HT TypingH2 TypingP2; subst...

  forwards : tx_store_invariant_step_many...
  forwards [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] : IHStepMany...
  forwards [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] :
    strong_lazy_to_inverted_lazy_step...
Qed.

Lemma strong_lazy_to_inverted_lazy : forall P0 H2 P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  LazySys.system_step_many Lazy.empty_heap P0 H2 P tr ->
  exists H3,
    map_heap_inverted_to_weak H3 H2
    /\ ILazySys.system_step_many ILazy.empty_heap P0 H3 P tr.
Proof with eauto.
  intros P0 H2 P tr TypingP0 StepMany.
  forwards TypingH0 : TypedStrongLazy.typing_empty_heap.
  forwards Invariant0 : tx_store_invariant_empty.
  forwards MapHeap0 : map_heap_inverted_to_weak_empty.
  forwards [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] :
    strong_lazy_to_inverted_lazy_many...
Qed.
