Require Import init_modules.
Require Import inverted_to_strong_map_heap.

Import Heap.

Inductive not_commit : inverted_heap -> Prop :=
| not_commit_none : forall IS3 R3,
  not_commit (tx_none, IS3, R3)
| not_commit_running : forall t IS3 R3,
  not_commit (tx_running t, IS3, R3).

Lemma not_commit_update : forall ts3 IS3 R3 IS3' R3',
  not_commit (ts3, IS3, R3) ->
  not_commit (ts3, IS3', R3').
Proof.
  intros ts3 IS3 R3 IS3' R3' NotCommit.
  inversion NotCommit.
  apply not_commit_none.
  apply not_commit_running.
Qed.

Lemma not_commit_active : forall ts3 IS3 R3 t,
  not_commit (ts3, IS3, R3) ->
  weak_active ts3 t.
Proof.
  intros ts3 IS3 R3 t NotCommit.
  destruct ts3; subst.
  apply weak_active_none.
  apply weak_active_running.
  inversion NotCommit.
Qed.

Lemma step_pure : forall H3 H1 H1' t,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, pure _)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, pure _)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t'' l' r' v' v'' RgnFresh Fresh Active Read
    NotRunning.
  esplit. split...
  apply IWeak.step_pure...
  eapply not_commit_active...
Qed.

Lemma step_newRgn : forall H3 H1 H1' t r,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, newRgn _ r)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, newRgn _ r)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t r MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r' v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  apply IWeak.step_newRgn...
  eapply not_commit_active...
  apply map_heap_inverted_to_strong_def...
  inversion NotCommit; subst.
  apply not_commit_none.
  apply not_commit_running.
Qed.

Lemma step_alloc : forall H3 H1 H1' t l r v,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, alloc l r v)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, alloc l r v)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t l r v MapHeap NotCommit
    Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t'' l' r' v' v'' RgnFresh Fresh Active Read
    NotRunning.
  esplit. split; [ | split ].
  apply IWeak.step_alloc...
  eapply not_commit_active...
  rewrite <- map_heap_fresh...
  apply map_heap_alloc...
  rewrite <- map_heap_fresh...
  destruct ts3; subst.
  apply not_commit_none.
  apply not_commit_running.
  inversion NotCommit.
Qed.

Lemma step_read : forall H3 H1 H1' t l v,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, read l v)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, read l v)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t l v MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l' r v' v'' RgnFresh Fresh Active Read
    NotRunning.

  (* what is stored in the inverted heap? *)
  rewrite VarMapFacts.find_mapsto_iff in Read.
  specializes MapStore l.
  rewrite Read in MapStore.
  remember (VarMap.find l IS3) as Read3.
  symmetry in HeqRead3.
  destruct Read3 as [ [ [ s r' ] v' ] | ];
    inversion MapStore. subst r' v'.

  (* is there a transaction active? *)
  inversion Active; subst.

  (* no active transaction *)
  destruct ts3.
  esplit. split; [ | split ].
  eapply IWeak.step_read.
  rewrite VarMapFacts.find_mapsto_iff...
  auto. auto.
  inversion MapTxState.
  inversion NotCommit.

  (* active transaction *)
  destruct ts3;
    inversion MapTxState. subst.
  destruct s as [ s | ].
  esplit. split; [ | split ].
  eapply IWeak.step_tx_read2...
  rewrite VarMapFacts.find_mapsto_iff...
  auto. auto.
  esplit. split; [ | split ].
  eapply IWeak.step_tx_read1...
  rewrite VarMapFacts.find_mapsto_iff...
  apply map_heap_tx_read1...
  rewrite VarMapFacts.find_mapsto_iff...
  apply not_commit_running.
Qed.

Lemma step_write : forall H3 H1 H1' t l v,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, write l v)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, write l v)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t l v MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l' r v' v'' RgnFresh Fresh Active Read
    NotRunning.

  (* what is stored in the inverted heap? *)
  rewrite VarMapFacts.find_mapsto_iff in Read.
  specializes MapStore l.
  rewrite Read in MapStore.
  remember (VarMap.find l IS3) as Read3.
  symmetry in HeqRead3.
  destruct Read3 as [ [ [ s r' ] v'' ] | ];
    inversion MapStore. subst r' v''.

  (* is there a transaction active? *)
  inversion Active; subst.

  (* no active transaction *)
  destruct ts3.
  esplit. split; [ | split ].
  eapply IWeak.step_write.
  rewrite VarMapFacts.find_mapsto_iff...
  eapply map_heap_write...
  rewrite VarMapFacts.find_mapsto_iff...
  eapply not_commit_update...
  inversion MapTxState.
  inversion NotCommit.

  (* active transaction *)
  destruct ts3;
    inversion MapTxState. subst.
  esplit. split; [ | split ].
  eapply IWeak.step_tx_write...
  rewrite VarMapFacts.find_mapsto_iff...
  eapply map_heap_write...
  rewrite VarMapFacts.find_mapsto_iff...
  apply not_commit_running.
Qed.

Lemma step_spawn : forall H3 H1 H1' t t',
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, spawn _ t')) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, spawn _ t')) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t t' MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t'' l r v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  apply IWeak.step_spawn.
  destruct ts3; subst.
  apply weak_unprotected_none.
  apply weak_unprotected_running.
  intro. subst. apply NotRunning...
  inversion NotCommit.
  auto. auto.
Qed.

Lemma step_makeP  : forall H3 H1 H1' t r,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, makeP _ r)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, makeP _ r)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t r MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r' v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  destruct ts3.
  apply IWeak.step_makeP.
  inversion MapTxState.
  inversion NotCommit.
  apply map_heap_inverted_to_strong_def...
  apply not_commit_none...
Qed.

Lemma step_makeU  : forall H3 H1 H1' t r,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, makeU _ r)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, makeU _ r)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t r MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r' v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  destruct ts3.
  apply IWeak.step_makeU.
  inversion MapTxState.
  inversion NotCommit.
  apply map_heap_inverted_to_strong_def...
  apply not_commit_none...
Qed.

Lemma step_makeRO  : forall H3 H1 H1' t r,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, makeRO _ r)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, makeRO _ r)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t r MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r' v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  destruct ts3.
  apply IWeak.step_makeRO.
  inversion MapTxState.
  inversion NotCommit.
  apply map_heap_inverted_to_strong_def...
  apply not_commit_none...
Qed.

Lemma step_makeTL  : forall H3 H1 H1' t r,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, makeTL _ r)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, makeTL _ r)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t r MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r' v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  destruct ts3.
  apply IWeak.step_makeTL.
  inversion MapTxState.
  inversion NotCommit.
  apply map_heap_inverted_to_strong_def...
  apply not_commit_none...
Qed.

Lemma step_beginP  : forall H3 H1 H1' t,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, beginP _)) H1' ->
  exists H3',
    IWeak.heap_step H3 (Some (t, beginP _)) H3'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r v v' RgnFresh Fresh Active Read NotRunning.
  esplit. split; [ | split ].
  destruct ts3.
  apply IWeak.step_beginP.
  inversion MapTxState.
  inversion NotCommit.
  apply map_heap_inverted_to_strong_def...
  apply not_commit_running...
Qed.

Definition countLogEntries (r : rgn) (o : option tx_loc_state * rgn * L.val)
  (n : nat) : nat :=
  let (q, _) := o in
    let (s, _) := q in
      match s with
        | None => n
        | Some _ => S n
      end.

Lemma proper_eq : Proper ((@eq _) ==> (@eq _) ==> (@eq _) ==> (@eq _))
  countLogEntries.
Proof.
  intros ? ? ? ? ? ? ? ? ?.
  subst. auto.
Qed.

Lemma transpose_neqkey_count :
  VarMapProperties.transpose_neqkey (@eq _) countLogEntries.
Proof.
  intros ? ? [ [ [ ? | ] ? ] ? ] [ [ [ ? | ] ? ] ? ] ? ?;
    unfold countLogEntries; auto.
Qed.

Hint Resolve proper_eq transpose_neqkey_count.

Lemma commit_all : forall IS3 R3 S1 R1 t,
  map_heap_inverted_to_strong (tx_commit t, IS3, R3) (tx_none, S1, R1) ->
  exist IS3' tr,
    IWeak.heap_step_many (tx_commit t, IS3, R3) (tx_commit t, IS3', R3) tr
    /\ erase_trace tr = nil
    /\ (forall l s r v, VarMap.MapsTo l (s, r, v) IS3' -> s = None)
    /\ map_heap_inverted_to_strong (tx_commit t, IS3', R3) (tx_none, S1, R1).
Proof with eauto.
  intros IS3 R3 S1 R1 t MapHeap.
  remember (VarMap.fold countLogEntries IS3 0) as n.
  symmetry in Heqn.
  generalize dependent IS3.
  induction n; intros IS3 MapHeap Heqn.

  (* Base case - the TX store is empty. *)

  (* This is a freaking mess! I'm adding some notes just to explain
     what's going on. *)
  exists IS3. exists (nil : trace L.val).
  split. apply IWeak.heap_step_many_refl.
  split... split...

  (* Assume that there is some entry in IS3 whose transaction state is
     not None. Then we'll show a contradiction by showing that
     countLogEntries must be greater than zero. *)
  intros l s r v Read3.
  destruct s as [ s | ]...

  (* Let IS3' be the store with the log entry removed. Let IS3'' be
     IS3' with the log entry added back in. *)
  remember (VarMap.remove l IS3) as IS3'.
  remember (VarMap.add l (Some s, r, v) IS3') as IS3''.

  (* IS3 and IS3'' have the same number of log entries. *)
  assert (VarMap.fold countLogEntries IS3 0 =
    VarMap.fold countLogEntries IS3'' 0) as Equality1.
  apply VarMapProperties.fold_Equal with (eqA := (@eq nat))...
  subst IS3' IS3''.
  intro l'.
  destruct (l == l').
  rewrite VarMapFacts.add_eq_o...
  subst.
  rewrite <- VarMapFacts.find_mapsto_iff...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.remove_neq_o...

  (* The number of log entries in IS3'' is the number in IS3' plus
     1. *)
  assert (VarMap.fold countLogEntries IS3'' 0 =
    countLogEntries l (Some s, r, v) (VarMap.fold countLogEntries IS3' 0))
  as Equality2.
  subst IS3' IS3''.
  apply VarMapProperties.fold_add with (eqA := (@eq nat))...
  apply VarMap.remove_1.
  apply Var_as_OT.eq_refl.

  (* Contradiction - we get that 0 is equal to something that's at
     least 1. *)
  rewrite Equality1 in Heqn.
  rewrite Equality2 in Heqn.
  unfold countLogEntries in *.
  inversion Heqn.

  (* Inductive case - The number of log entries in IS3 is at least 1 -
     need to show that such an entry exists and remove it! *)
  rewrite VarMap.fold_1 in Heqn.
  remember (VarMap.elements IS3) as IS3_list.
  assert (exist l s r v,
    InA (@VarMap.eq_key_elt _) (l, (Some s, r, v)) IS3_list)
    as [ l [ s [ r [ v LogEntry ] ] ] ].
  clear HeqIS3_list.
  induction IS3_list as [ | [ l p ] IS3_list ].
  discriminate.
  destruct p as [ [ [ ? | ] ? ] ? ].
  repeat esplit.
  apply InA_cons_hd.
  unfold VarMap.eq_key_elt.
  split...
  apply Var_as_OT.eq_refl.
  unfold countLogEntries in *.
  simpl in Heqn.
  specialize (IHIS3_list Heqn).
  destruct IHIS3_list as [ ? [ ? [ ? [ ? ? ] ] ] ].
  repeat esplit.
  apply InA_cons_tl...
  subst IS3_list.
  rewrite <- VarMapFacts.elements_mapsto_iff in LogEntry.
  rewrite <- VarMap.fold_1 in Heqn.

  (* OK, the log entry exists - now show that removing it decreases
     the log count by 1. *)
  remember (VarMap.remove l IS3) as IS3'.
  remember (VarMap.add l (Some s, r, v) IS3') as IS3''.
  remember (VarMap.add l (None, r, v) IS3) as IS3'''.
  remember (VarMap.add l (None, r, v) IS3') as IS3''''.

  (* IS3 = IS3'' *)
  assert (VarMap.fold countLogEntries IS3 0 =
    VarMap.fold countLogEntries IS3'' 0) as Equality1.
  apply VarMapProperties.fold_Equal with (eqA := (@eq nat))...
  subst IS3' IS3''.
  intros l'.
  destruct (l == l').
  rewrite VarMapFacts.add_eq_o...
  rewrite <- VarMapFacts.find_mapsto_iff.
  subst...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.remove_neq_o...

  (* IS3'' = IS3' + 1 *)
  assert (VarMap.fold countLogEntries IS3'' 0 =
    countLogEntries l (Some s, r, v) (VarMap.fold countLogEntries IS3' 0))
  as Equality2.
  subst IS3''.
  apply VarMapProperties.fold_add with (eqA := (@eq nat))...
  subst IS3'.
  apply VarMap.remove_1.
  apply Var_as_OT.eq_refl.

  (* IS3''' = IS3'''' *)
  assert (VarMap.fold countLogEntries IS3''' 0 =
    VarMap.fold countLogEntries IS3'''' 0) as Equality3.
  apply VarMapProperties.fold_Equal with (eqA := (@eq nat))...
  subst IS3' IS3''' IS3''''.
  intros l'.
  destruct (l == l').
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_eq_o...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.add_neq_o...
  rewrite VarMapFacts.remove_neq_o...

  (* IS3'''' = IS3' + 0 *)
  assert (VarMap.fold countLogEntries IS3'''' 0 =
    countLogEntries l (None, r, v) (VarMap.fold countLogEntries IS3' 0))
  as Equality4.
  subst IS3' IS3''''.
  apply VarMapProperties.fold_add with (eqA := (@eq nat))...
  apply VarMap.remove_1.
  apply Var_as_OT.eq_refl.

  (* use these four equalities to instantiate the inductive
     hypothesis. *)
  rewrite Equality1 in Heqn.
  rewrite Equality2 in Heqn.
  unfold countLogEntries in *. simpl in *.
  rewrite <- Equality4 in Heqn.
  rewrite <- Equality3 in Heqn.
  inversion Heqn as [ Equality5 ].
  specializes IHn (>>> IS3''' __ __).
  subst IS3'''.
  eapply map_heap_commit...
  apply Equality5.

  (* use the witness provided by the induction hypothesis as our
     witness. *)
  subst IS3'''.
  destruct IHn as [ IS3''''' [ ? [ ? [ ? [ ? ? ] ] ] ] ].
  exists IS3'''''. esplit. split.
  eapply IWeakFacts.heap_step_many_alt.
  eapply IWeak.heap_step_many_trans.
  apply IWeak.heap_step_many_refl.
  eapply IWeak.step_commit...
  eauto.
  split...
  rewrite app_erase.
  simpl...
  rewrite <- app_nil_end...
Qed.

Lemma step_endP  : forall H3 H1 H1' t,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 (Some (t, endP _)) H1' ->
  exists H3', exists tr,
    IWeak.heap_step_many H3 H3' tr
    /\ erase_trace tr = (t, endP _) :: nil
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto.
  intros [ [ ts3 IS3 ] R3 ] [ [ ts1 S1 ] R1 ] H1' t MapHeap NotCommit Step.
  inversion_map_heap MapHeap MapTxState MapStore.
  inversion_strong_step Step t' l r v v' RgnFresh Fresh Active Read NotRunning.
  destruct ts3 as [ | t' | ]; inversion MapTxState; subst t'.
  forwards MapHeap' : map_heap_endP (>>> __)...
  forwards [ ? [ ? [ ? [ EraseEq [ ? ? ] ] ] ] ] : commit_all (>>> __)...
  esplit. esplit. split; [ | split; [ | split ] ].
  eapply IWeak.heap_step_many_trans.
  eapply IWeakFacts.heap_step_many_alt.
  eapply IWeak.heap_step_many_trans.
  apply IWeak.heap_step_many_refl.
  apply IWeak.step_endP.
  eauto.
  apply IWeak.step_end_commit.
  eauto.
  simpl.
  rewrite app_erase.
  rewrite EraseEq.
  simpl...
  eapply map_heap_end_commit...
  apply not_commit_none.
Qed.

Hint Constructors IWeak.heap_step_many.

Lemma strong_to_inverted_weak_step  : forall H3 H1 H1' e,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  Strong.heap_step H1 e H1' ->
  exists H3', exists tr,
    erase_trace tr = erase_trace (e :: nil)
    /\ IWeak.heap_step_many H3 H3' tr
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto 7.
  intros H3 H1 H1' e MapHeap NotCommit Step.
  destruct e as [ [ t a ] | ].
  destruct a.
  forwards [ ? [ ? [ ? ? ] ] ] : step_pure...
  forwards [ ? [ ? [ ? ? ] ] ] : step_newRgn...
  forwards [ ? [ ? [ ? ? ] ] ] : step_alloc...
  forwards [ ? [ ? [ ? ? ] ] ] : step_read...
  forwards [ ? [ ? [ ? ? ] ] ] : step_write...
  forwards [ ? [ ? [ ? ? ] ] ] : step_spawn...
  forwards [ ? [ ? [ ? ? ] ] ] : step_makeP...
  forwards [ ? [ ? [ ? ? ] ] ] : step_makeU...
  forwards [ ? [ ? [ ? ? ] ] ] : step_makeRO...
  forwards [ ? [ ? [ ? ? ] ] ] : step_makeTL...
  forwards [ ? [ ? [ ? ? ] ] ] : step_beginP...
  forwards [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] : step_endP...
  inversion Step.
Qed.

Lemma strong_to_inverted_weak_many : forall H3 H1 P H1' P' tr,
  map_heap_inverted_to_strong H3 H1 ->
  not_commit H3 ->
  StrongSys.system_step_many H1 P H1' P' tr ->
  exists H3', exists tr',
    IWeakSys.system_step_many H3 P H3' P' tr'
    /\ erase_trace tr = erase_trace tr'
    /\ map_heap_inverted_to_strong H3' H1'
    /\ not_commit H3'.
Proof with eauto 6.
  intros H3 H1 P H1' P' tr MapHeap NotCommit StepMany.
  induction StepMany as [ | ? ? ? ? ? ? ? ? ? ? Step ]...
  specializes IHStepMany (>>> __)...
  destruct IHStepMany as [ ? [ ? [ ? [ EraseEq1 [ ? ? ] ] ] ] ].
  inversion Step as [ | ? ? ? HeapStep ]; subst.
  forwards [ ? [ ? [ EraseEq2 [ ? [ ? ? ] ] ] ] ]
  : strong_to_inverted_weak_step (>>> __ __ __)...
  esplit. esplit. split.
  eapply IWeakSysFacts.system_step_many_alt...
  eapply IWeakSysFacts.system_step_many_mutual...
  split...
  simpl in *.
  rewrite EraseEq1.
  rewrite app_erase.
  rewrite EraseEq2...
  inversion HeapStep.
Qed.

Lemma strong_to_inverted_weak : forall P H1 P' tr,
  StrongSys.system_step_many Strong.empty_heap P H1 P' tr ->
  exists H3, exists tr',
    IWeakSys.system_step_many IWeak.empty_heap P H3 P' tr'.
Proof with eauto.
  intros P H1 P' tr StepMany.
  forwards [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] : strong_to_inverted_weak_many...
  apply map_heap_empty.
  apply not_commit_none.
Qed.
