Require Import heaps.
Require Import maps.
Require Import types.

Module TypedStrongHeap
  (Typ : TYPE_MODEL with Module HeapTyp := HeapTyping)
  (Heap : HEAP Typ.L).

  Inductive typing_heap : HeapTyping.heap_typing Typ.typ -> Heap.strong_heap
    -> Prop :=
  | typing_heap_def : forall PT ST ts S R,
    (forall l,
      VarMap.In l ST <-> VarMap.In l S) ->
    (forall l r v T,
      VarMap.MapsTo l (r, v) S ->
      VarMap.MapsTo l T ST ->
      Typ.typing_val (PT, ST) v T) ->
    (forall l r v,
      VarMap.MapsTo l (r, v) S ->
      VarMap.In r PT) ->
    (forall r,
      VarMap.In r PT <-> VarMap.In r R) ->
    typing_heap (PT, ST) (ts, S, R).

  Lemma typing_empty_heap :
    typing_heap (Typ.HeapTyp.empty_heap_typing _) Heap.empty_strong_heap.
  Proof with auto.
    apply typing_heap_def.

    intros.
    repeat rewrite VarMapFacts.empty_in_iff.
    split; intro false; inversion false.

    intros ? ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros.
    repeat rewrite VarMapFacts.empty_in_iff.
    split; intro false; inversion false.
  Qed.
End TypedStrongHeap.

Ltac inversion_typing_strong_heap TypingH DomStore TypingStore RgnStore DomProt
  :=
  inversion TypingH as [ ? ? ? ? ? DomStore TypingStore RgnStore DomProt ];
    subst.

Module TypedInvertedHeap
  (Typ : TYPE_MODEL with Module HeapTyp := HeapTyping)
  (Heap : HEAP Typ.L).

  Inductive typing_heap : HeapTyping.heap_typing Typ.typ -> Heap.inverted_heap
    -> Prop :=
  | typing_heap_def : forall PT ST ts IS R,
    (forall l,
      VarMap.In l ST <-> VarMap.In l IS) ->
    (forall l s r v T,
      VarMap.MapsTo l (s, r, v) IS ->
      VarMap.MapsTo l T ST ->
      Typ.typing_val (PT, ST) v T) ->
    (forall l s r v,
      VarMap.MapsTo l (s, r, v) IS ->
      VarMap.In r PT) ->
    (forall r,
      VarMap.In r PT <-> VarMap.In r R) ->
    typing_heap (PT, ST) (ts, IS, R).

  Lemma typing_empty_heap :
    typing_heap (Typ.HeapTyp.empty_heap_typing _) Heap.empty_inverted_heap.
  Proof with auto.
    apply typing_heap_def.

    intros.
    repeat rewrite VarMapFacts.empty_in_iff.
    split; intro false; inversion false.

    intros ? ? ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros ? ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros.
    repeat rewrite VarMapFacts.empty_in_iff.
    split; intro false; inversion false.
  Qed.
End TypedInvertedHeap.

Ltac inversion_typing_inverted_heap TypingH DomStore TypingStore RgnStore
  DomProt :=
  inversion TypingH as [ ? ? ? ? ? DomStore TypingStore RgnStore DomProt ];
    subst.

Module TypedWeakHeap
  (Typ : TYPE_MODEL with Module HeapTyp := HeapTyping)
  (Heap : HEAP Typ.L).

  Inductive typing_heap : HeapTyping.heap_typing Typ.typ -> Heap.weak_heap
    -> Prop :=
  | typing_heap_def : forall PT ST ts S R TS,
    (forall l,
      VarMap.In l ST <-> VarMap.In l S) ->
    (forall l,
      VarMap.In l TS ->
      VarMap.In l ST) ->
    (forall l r v T,
      VarMap.MapsTo l (r, v) S ->
      VarMap.MapsTo l T ST ->
      Typ.typing_val (PT, ST) v T) ->
    (forall l r v,
      VarMap.MapsTo l (r, v) S ->
      VarMap.In r PT) ->
    (forall l s v T,
      VarMap.MapsTo l (s, v) TS ->
      VarMap.MapsTo l T ST ->
      Typ.typing_val (PT, ST) v T) ->
    (forall r,
      VarMap.In r PT <-> VarMap.In r R) ->
    typing_heap (PT, ST) (ts, S, R, TS).

  Lemma typing_empty_heap :
    typing_heap (Typ.HeapTyp.empty_heap_typing _) Heap.empty_weak_heap.
  Proof with auto.
    apply typing_heap_def.

    intros.
    repeat rewrite VarMapFacts.empty_in_iff.
    split; intro false; inversion false.

    intros ? false.
    rewrite VarMapFacts.empty_in_iff in false.
    inversion false.

    intros ? ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros ? ? ? ? false.
    rewrite VarMapFacts.empty_mapsto_iff in false.
    inversion false.

    intros.
    repeat rewrite VarMapFacts.empty_in_iff.
    split; intro false; inversion false.
  Qed.
End TypedWeakHeap.

Ltac inversion_typing_weak_heap TypingH DomStore DomTx TypingStore RgnStore
  TypingTx DomProt :=
  inversion TypingH as [ ? ? ? ? ? ? DomStore DomTx TypingStore RgnStore
    TypingTx DomProt ]; subst.
