Require Import init_modules.

Module TypedIWeak <: TYPED_MEMORY_MODEL Typ.
  Module M := IWeak.
  Module TypedInvertedHeap := TypedInvertedHeap Typ Heap.
  Definition typing_heap := TypedInvertedHeap.typing_heap.
  Definition typing_empty_heap := TypedInvertedHeap.typing_empty_heap.
End TypedIWeak.

Module TypedIWeakPreservation <: HEAP_TYPE_PRESERVATION Typ.
  Module TypM := TypedIWeak.

  Hint Unfold Typ.well_typed_heap_action.
  Hint Constructors Typ.well_typed_heap_action_ind.
  Hint Resolve Typ.typing_stability.

  Lemma well_typed_heap_step_mutual : forall HT H t a H',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H (Some (t, a)) H' ->
    Typ.well_typed_heap_action HT a.
  Proof with eauto.
    intros HT H t a H' TypingH Step.
    destruct HT as [ PT ST ].
    destruct H as [ [ ts IS ] R ].
    inversion_typing_inverted_heap TypingH DomStore TypingStore RgnStore
      DomProt.
    inversion_inverted_weak_step Step t' l s r v v' Active RgnFresh Fresh Read
      Unprotected Empty...

    (* newRgn *)
    apply Typ.well_typed_heap_action_newRgn.
    rewrite DomProt...

    (* alloc *)
    apply Typ.well_typed_heap_action_alloc.
    rewrite DomStore...

    (* tx_read1 *)
    assert (VarMap.In l ST) as [ ? ? ].
    rewrite DomStore...
    eapply Typ.well_typed_heap_action_read...

    (* tx_read2 *)
    assert (VarMap.In l ST) as [ ? ? ].
    rewrite DomStore...
    eapply Typ.well_typed_heap_action_read...

    (* read *)
    assert (VarMap.In l ST) as [ ? ? ].
    rewrite DomStore...
    eapply Typ.well_typed_heap_action_read...
  Qed.

  Lemma type_preservation_heap_step_mutual : forall HT H t a H' HT',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H (Some (t, a)) H' ->
    Typ.well_typed_program_action HT a HT' ->
    TypM.typing_heap HT' H'.
  Proof with eauto.
    intros HT H t a H' HT' TypingH Step TypingA.
    destruct HT as [ PT ST ].
    destruct H as [ [ ts IS ] R ].
    forwards : Typ.well_typed_extends...
    eapply well_typed_heap_step_mutual...
    inversion_typing_inverted_heap TypingH DomStore TypingStore RgnStore
      DomProt.
    inversion_inverted_weak_step Step t' l s r v v' Active RgnFresh Fresh Read
      Unprotected Empty;
      inversion_well_typed_program_action TypingA TypingRgn TypingVal TypingLoc;
        subst; eauto;
          apply TypM.TypedInvertedHeap.typing_heap_def...

    (* newRgn *)
    intros l s r' v Read.
    rewrite VarMapFacts.add_in_iff...
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomProt.
    split...

    (* alloc *)
    intros l'.
    rewrite VarMapFacts.add_in_iff.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros l' s r' v' T' Read TypingLoc.
    eapply Typ.typing_stability...
    rewrite VarMapFacts.add_mapsto_iff in Read.
    destruct Read as [ [ ? Eq ] | [ ? ? ] ].
    inversion Eq; subst s r' v'.
    rewrite VarMapFactsEx2.add_eq_mapsto_iff in TypingLoc...
    subst...
    rewrite VarMapFacts.add_neq_mapsto_iff in TypingLoc...
    intros l' s r' v' Read.
    rewrite VarMapFacts.add_mapsto_iff in Read.
    destruct Read as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...

    (* tx_read1 *)
    intros l'.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros [ ? | ? ]...
    subst...
    intros l' s r' v' T Read' TypingLoc.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    intros l' s r' v' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...

    (* tx_write *)
    intros l'.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros [ ? | ? ]...
    subst...
    intros l' s' r' v'' T' Read' TypingLoc'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst.
    assert (T = T').
    eapply VarMapFacts.MapsTo_fun...
    subst...
    intros l' s' r' v'' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...

    (* write *)
    intros l'.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros [ ? | ? ]...
    subst...
    intros l' s' r' v'' T' Read' TypingLoc'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst.
    assert (T = T').
    eapply VarMapFacts.MapsTo_fun...
    subst...
    intros l' s' r' v'' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...

    (* makeP *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...

    (* makeU *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...

    (* makeRO *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...

    (* makeTL *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...
  Qed.

  Lemma type_preservation_heap_step_empty : forall HT H H',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H None H' ->
    TypM.typing_heap HT H'.
  Proof with eauto.
    intros HT H H' TypingH Step.
    destruct HT as [ PT ST ].
    destruct H as [ [ ts IS ] R ].
    inversion_typing_inverted_heap TypingH DomStore TypingStore RgnStore
      DomProt.
    inversion_inverted_weak_step Step t l s r v v' Active RgnFresh Fresh Read
      Unprotected Empty; eauto;
        apply TypM.TypedInvertedHeap.typing_heap_def...

    (* commit_ro *)
    intros l'.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros [ ? | ? ]...
    subst...
    intros l' s' r' v' T Read' TypingLoc.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    intros l' s' r' v' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
  Qed.
End TypedIWeakPreservation.

Module IWeakSysPreservation := Type_Preservation HeapTyping Typ PrgmPres
  TypedIWeakPreservation IWeakSys.
