Require Import init_modules.

Module TypedStrongLazy <: TYPED_MEMORY_MODEL Typ.
  Module M := Lazy.
  Module TypedWeakHeap := TypedWeakHeap Typ Heap.
  Definition typing_heap := TypedWeakHeap.typing_heap.
  Definition typing_empty_heap := TypedWeakHeap.typing_empty_heap.
End TypedStrongLazy.

Module TypedStrongLazyPreservation <: HEAP_TYPE_PRESERVATION Typ.
  Module TypM := TypedStrongLazy.

  Hint Unfold Typ.well_typed_heap_action.
  Hint Constructors Typ.well_typed_heap_action_ind.
  Hint Resolve Typ.typing_stability.

  Lemma well_typed_heap_step_mutual : forall HT H t a H',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H (Some (t, a)) H' ->
    Typ.well_typed_heap_action HT a.
  Proof with eauto.
    intros ? ? ? ? ? TypingH Step.
    destruct HT as [ PT ST ].
    destruct H as [ [ [ ts S ] R ] TS ].
    inversion_typing_weak_heap TypingH DomStore DomTx TypingStore RgnStore
      TypingTx DomProt.
    inversion_lazy_step Step t' l r v v' Active RgnFresh Fresh Read TxFresh
      TxRead Empty...

    (* newRgn *)
    apply Typ.well_typed_heap_action_newRgn...
    rewrite DomProt...

    (* alloc *)
    apply Typ.well_typed_heap_action_alloc.
    rewrite DomStore...

    (* tx_read1 *)
    assert (VarMap.In l S) as InStore...
    rewrite <- DomStore in InStore.
    destruct InStore as [ ? ? ].
    eapply Typ.well_typed_heap_action_read...

    (* tx_read2 *)
    assert (VarMap.In l TS) as InTx...
    apply DomTx in InTx.
    destruct InTx as [ ? ? ].
    eapply Typ.well_typed_heap_action_read...

    (* read *)
    assert (VarMap.In l S) as InStore...
    rewrite <- DomStore in InStore.
    destruct InStore as [ ? ? ].
    eapply Typ.well_typed_heap_action_read...
  Qed.

  Lemma type_preservation_heap_step_mutual : forall HT H t a H' HT',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H (Some (t, a)) H' ->
    Typ.well_typed_program_action HT a HT' ->
    TypM.typing_heap HT' H'.
  Proof with eauto.
    intros HT H t a H' HT' TypingH Step TypingA.
    forwards Extends : Typ.well_typed_extends...
    eapply well_typed_heap_step_mutual...
    inversion_typing_weak_heap TypingH DomStore DomTx TypingStore RgnStore
      TypingTx DomProt.
    inversion_lazy_step Step t' l r v v' Active RgnFresh Fresh Read TxFresh
      TxRead Empty;
      inversion_well_typed_program_action TypingA TypingRgn TypingVal TypingLoc;
        subst; eauto;
          apply TypM.TypedWeakHeap.typing_heap_def...

    (* newRgn *)
    intros l r' v Read.
    rewrite VarMapFacts.add_in_iff...
    intros r'.
    repeat rewrite VarMapFacts.add_in_iff.
    rewrite DomProt.
    split...

    (* alloc *)
    intros l'.
    repeat rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros.
    rewrite VarMapFacts.add_in_iff...
    intros l' r' v' T' Read' TypingVal'.
    eapply Typ.typing_stability...
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ].
    inversion Eq; subst r' v'.
    rewrite VarMapFactsEx2.add_eq_mapsto_iff in TypingVal'...
    subst...
    rewrite VarMapFacts.add_neq_mapsto_iff in TypingVal'...
    intros l' r' v' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    intros l' s v' T' TxRead TypingLoc.
    rewrite VarMapFacts.add_neq_mapsto_iff in TypingLoc...
    intro. subst.
    apply Fresh.
    rewrite <- DomStore...

    (* tx_read1 *)
    intros l' TxRead.
    rewrite VarMapFacts.add_in_iff in TxRead.
    destruct TxRead...
    subst.
    rewrite DomStore...
    intros l' s v' T TxRead TypingLoc.
    rewrite VarMapFacts.add_mapsto_iff in TxRead.
    destruct TxRead as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...

    (* tx_write *)
    intros l' TxRead.
    rewrite VarMapFacts.add_in_iff in TxRead.
    destruct TxRead...
    subst...
    intros l' s v' T' TxRead TypingLoc'.
    rewrite VarMapFacts.add_mapsto_iff in TxRead.
    destruct TxRead as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    assert (T = T').
    eapply VarMapFacts.MapsTo_fun...
    subst...

    (* write *)
    intros l'.
    repeat rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros [ ? | ? ]...
    subst...
    intros l' r' v'' T' Read' TypingLoc'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    assert (T = T').
    eapply VarMapFacts.MapsTo_fun...
    subst...
    intros l' r' v'' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...

    (* makeP *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...

    (* makeU *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...

    (* makeRO *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...

    (* makeTL *)
    intros r'.
    rewrite VarMapFacts.add_in_iff.
    rewrite <- DomProt.
    split...
    intros [ ? | ? ]...
    subst...
  Qed.

  Lemma type_preservation_heap_step_empty : forall HT H H',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H None H' ->
    TypM.typing_heap HT H'.
  Proof with eauto.
    intros HT H H' TypingH Step.
    destruct HT as [ PT ST ].
    destruct H as [ [ [ ts S ] R ] TS ].
    inversion_typing_weak_heap TypingH DomStore DomTx TypingStore RgnStore
      TypingTx DomProt.
    inversion_lazy_step Step t' l r v v' Active RgnFresh Fresh Read TxFresh
      TxRead Empty; eauto;
        apply TypM.TypedWeakHeap.typing_heap_def...

    (* commit_ro *)
    intros l' InTx.
    rewrite VarMapFacts.remove_in_iff in InTx.
    destruct InTx...
    intros l' s v' T TxRead' TypingLoc'.
    rewrite VarMapFacts.remove_mapsto_iff in TxRead'.
    destruct TxRead'...

    (* commit_mod *)
    intros l'.
    rewrite VarMapFacts.add_in_iff.
    rewrite DomStore.
    split...
    intros [ ? | ? ]...
    subst...
    intros l' InTx.
    rewrite VarMapFacts.remove_in_iff in InTx.
    destruct InTx...
    intros l' r' v'' T Read' TypingLoc'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    intros l' r' v'' Read'.
    rewrite VarMapFacts.add_mapsto_iff in Read'.
    destruct Read' as [ [ ? Eq ] | [ ? ? ] ]...
    inversion Eq; subst...
    intros l' s v'' T TxRead' TypingLoc.
    rewrite VarMapFacts.remove_mapsto_iff in TxRead'.
    destruct TxRead'...
  Qed.
End TypedStrongLazyPreservation.

Module LazySysPreservation := Type_Preservation HeapTyping Typ
  PrgmPres TypedStrongLazyPreservation LazySys.
