Require Import init_modules.
Require Import separation.
Require Import separation_lazy.
Require Import strong_lazy_to_inverted_lazy.
Require Import typed_strong_lazy.

Import Heap.

Set Implicit Arguments.

Inductive weak_heap_invariant : Weak.heap -> Prop :=
| weak_heap_invariant_def : forall ts S R TS,
  (forall l r v v' p,
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo l (tx_ro, v') TS ->
    VarMap.MapsTo r p R ->
    match ts with
      | tx_none => True
      | tx_running t
      | tx_commit t =>
        check mode_protected p t op_read
    end) ->
  (forall l r v v' p,
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo l (tx_mod, v') TS ->
    VarMap.MapsTo r p R ->
    match ts with
      | tx_none => True
      | tx_running t
      | tx_commit t =>
        check mode_protected p t op_write
    end) ->
  weak_heap_invariant (ts, S, R, TS).

Lemma weak_heap_invariant_empty :
  weak_heap_invariant Weak.empty_heap.
Proof with eauto.
  apply weak_heap_invariant_def...
Qed.

Lemma weak_heap_invariant_step : forall HT H P e H' P',
  Typ.typing_prgm HT P ->
  TypedStrongLazy.typing_heap HT H ->
  tx_store_invariant H ->
  separation_lazy H P ->
  weak_heap_invariant H ->
  LazySys.system_step H P e H' P' ->
  weak_heap_invariant H'.
Proof with eauto.
  intros [ PT ST ] H P e H' P' TypingPrgm TypingHeap TxInvariant Separation
    Invariant Step.
  inversion_typing_weak_heap TypingHeap DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  inversion Invariant as [ ? ? ? ? ReadActive WriteActive ]; subst.
  inversion Step as [ ? ? ? ? t ? ProgramStep HeapStep | ? ? ? HeapStep ];
    subst;
  inversion_lazy_step HeapStep t' l r v v' Active RgnFresh Fresh Read TxFresh
    TxRead Empty...

  (* newRgn *)
  apply weak_heap_invariant_def.
  intros l r' v v' p Read TxRead Prot.
  rewrite VarMapFacts.add_neq_mapsto_iff in Prot.
  eapply ReadActive...
  intro. subst. apply RgnFresh. rewrite <- DomProt...
  intros l r' v v' p Read TxRead Prot.
  rewrite VarMapFacts.add_neq_mapsto_iff in Prot.
  eapply WriteActive...
  intro. subst. apply RgnFresh. rewrite <- DomProt...

  (* alloc *)
  apply weak_heap_invariant_def.
  intros l' r' v' v'' p Read TxRead Prot.
  rewrite VarMapFacts.add_neq_mapsto_iff in Read.
  eapply ReadActive...
  intro. subst. apply Fresh. rewrite <- DomStore...
  intros l' r' v' v'' p Read TxRead Prot.
  rewrite VarMapFacts.add_neq_mapsto_iff in Read.
  eapply WriteActive...
  intro. subst. apply Fresh. rewrite <- DomStore...

  (* tx_read1 *)
  apply weak_heap_invariant_def.
  intros l' r' v' v'' p Read' TxRead Prot.
  rewrite VarMapFacts.add_mapsto_iff in TxRead.
  inversion TxRead as [ [ ? Eq ] | [ ? ? ] ].
  inversion Eq; subst.
  forwards Eq' : VarMapFacts.MapsTo_fun (>>> Read Read').
  inversion Eq'; subst.
  inversion Separation; subst...
  eapply ReadActive...
  intros l' r' v' v'' p Read' TxRead Prot.
  rewrite VarMapFacts.add_mapsto_iff in TxRead.
  inversion TxRead as [ [ ? ? ] | [ ? ? ] ].
  discriminate.
  eapply WriteActive...

  (* tx_write *)
  apply weak_heap_invariant_def.
  intros l' r' v' v'' p Read' TxRead Prot.
  rewrite VarMapFacts.add_mapsto_iff in TxRead.
  inversion TxRead as [ [ ? ? ] | [ ? ? ] ].
  discriminate.
  eapply ReadActive...
  intros l' r v' v'' p Read TxRead Prot.
  rewrite VarMapFacts.add_mapsto_iff in TxRead.
  inversion TxRead as [ [ Eq ? ] | [ ? ? ] ].
  inversion Eq; subst.
  inversion Separation; subst...
  eapply WriteActive...

  (* write *)
  apply weak_heap_invariant_def...

  (* makeX *)
  apply weak_heap_invariant_def...
  apply weak_heap_invariant_def...
  apply weak_heap_invariant_def...
  apply weak_heap_invariant_def...

  (* beginP *)
  inversion TxInvariant as [ ? ? ? TxFresh | | ]; subst.
  apply weak_heap_invariant_def.
  intros. specializes TxFresh (>>> __).
  contradiction TxFresh...
  intros. specializes TxFresh (>>> __).
  contradiction TxFresh...

  (* endP *)
  apply weak_heap_invariant_def...

  (* commit_ro *)
  apply weak_heap_invariant_def...
  intros l' r v' v'' p Read' TxRead' Prot.
  rewrite VarMapFacts.remove_mapsto_iff in TxRead'.
  destruct TxRead'.
  eapply ReadActive...
  intros l' r v' v'' p Read' TxRead' Prot.
  rewrite VarMapFacts.remove_mapsto_iff in TxRead'.
  destruct TxRead'.
  eapply WriteActive...

  (* commit_mod *)
  apply weak_heap_invariant_def...
  intros l' r' v'' v''' p Read' TxRead' Prot.
  rewrite VarMapFacts.remove_mapsto_iff in TxRead'.
  destruct TxRead'.
  rewrite VarMapFacts.add_neq_mapsto_iff in Read'...
  intros l' r' v'' v''' p Read' TxRead' Prot.
  rewrite VarMapFacts.remove_mapsto_iff in TxRead'.
  destruct TxRead'.
  rewrite VarMapFacts.add_neq_mapsto_iff in Read'...

  (* end_commit *)
  apply weak_heap_invariant_def...
Qed.

Lemma weak_heap_invariant_step_many : forall P0 H P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  separation P0 ->
  LazySys.system_step_many Lazy.empty_heap P0 H P tr ->
  weak_heap_invariant H.
Proof with eauto.
  intros P0 H P tr TypingPrgm0 Separation0 LazySysStepMany.

  unfold Lazy.empty_heap in *.
  remember empty_weak_heap as H0.
  remember (HeapTyping.empty_heap_typing Typ.typ) as HT.
  generalize dependent HT.
  induction LazySysStepMany; intros HT HeqHT TypingPrgm.

  subst. apply weak_heap_invariant_empty.

  specializes IHLazySysStepMany (>>> __ __ __ __ __)...
  subst.
  forwards [ ? [ ? ? ] ] : LazySysPreservation.type_preservation_many...
  eapply weak_heap_invariant_step...
  eapply tx_store_invariant_step_many...
  apply tx_store_invariant_empty.
  eapply separation_to_separation_lazy...
Qed.
