Require Import init_modules.
Require Import list_facts.
Require Import separation.
Require Import separation_lazy.
Require Import separation_weak.
Require Import strong_lazy_to_inverted_lazy.
Require Import strong_lazy_to_weak.
Require Import weak_heap_invariant.
Require Import typed_strong_lazy.

Set Implicit Arguments.

Import Heap.

Hint Constructors Weak.step Lazy.step WeakSys.system_step LazySys.system_step
  Heap.strong_active Heap.weak_active Heap.weak_unprotected eq_weak_heap.
Hint Unfold Weak.heap_step Lazy.heap_step.
Hint Resolve L.eq_program_state_refl eq_weak_heap_refl tx_store_invariant_empty
  weak_heap_invariant_empty VarMapFacts.add_m.

Ltac not_program_order :=
  intro ProgramOrder; inversion ProgramOrder; subst; tauto.

Ltac reorder_program_steps :=
  forwards [ P2'' [ P3'' [ PrgmEq3' [ ProgramStep2'' ProgramStep3'' ] ] ] ]
  : LangReqs.program_reordering;
  [ eauto | eauto | not_program_order | ].

Lemma system_step_eq : forall H1 P1 e H2 P2 H1' P1',
  LazySys.system_step H1 P1 e H2 P2 ->
  L.eq_program_state P1 P1' ->
  eq_weak_heap H1 H1' ->
  exists H2', exists P2',
    L.eq_program_state P2 P2'
    /\ eq_weak_heap H2 H2'
    /\ LazySys.system_step H1' P1' e H2' P2'.
Proof with eauto.
  intros H1 P1 e H2 P2 H1' P1' Step PrgmEq HeapEq.
  inversion Step; subst.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step...
  forwards [ ? [ ? ? ] ] : Lazy.eq_weak_heap_step...
  eauto 6.

  forwards [ ? [ ? ? ] ] : Lazy.eq_weak_heap_step...
  eauto 6.
Qed.

Lemma weak_to_strong_lazy_none : forall S1 R1 TS1 P1 e1 H2 P2,
  WeakSys.system_step (tx_none, S1, R1, TS1) P1 e1 H2 P2 ->
  LazySys.system_step (tx_none, S1, R1, TS1) P1 e1 H2 P2.
Proof with eauto.
  intros S1 R1 TS1 P1 e1 H2 P2 WeakSysStep.
  inversion WeakSysStep as [ ? ? ? ? ? ? ? HeapStep | ? ? ? HeapStep ];
    inversion HeapStep; subst...
Qed.

Lemma reordering_pure : forall H1 P1 e1 H2 P2 H2' P2' t H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, pure _)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, pure _)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, pure _)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 9.
  intros [ [ [ ts1 S1 ] R1 ] TS1 ] P1 e1 [ [ [ ts2 S2 ] R2 ] TS2 ] P2
    [ [ [ ts2' S2' ] R2' ] TS2' ] P2' t H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ProgramStep3 WeakStep3 | ]; subst.
  forwards [ P3' [ PrgmEq3 ProgramStep3' ] ] : L.eq_program_state_step...
  inversion_weak_step WeakStep3 t' l r v v' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Unprotected2 Empty2.
  inversion Active2 as [ | t' | t' ]; subst;
    inversion HeapEq2; subst...

  (* running *)
  destruct (t == t').
  subst t'...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
    inversion_lazy_step LazyStep2' t'' l r v v' Active1 RgnFresh1 Fresh1 Read1
      TxFresh1 TxRead1 Empty1;
      try (inversion Active1; subst);
        reorder_program_steps;
          esplit; esplit; eauto 12 using L.eq_program_state_trans.

  (* commit *)
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2 LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
    inversion_lazy_step LazyStep2' t'' l r v v' Active1 RgnFresh1 Fresh1 Read1
    TxFresh1 TxRead1 Empty1; eauto 10;
      try (inversion Active1; subst);
        try reorder_program_steps;
        esplit; esplit; eauto 12 using L.eq_program_state_trans.
Qed.

Lemma reordering_strong_active_helper : forall H1 P1 H1' P1' t a H2 P2,
  WeakSys.system_step H1 P1 (Some (t, a)) H2 P2 ->
  L.eq_program_state P1 P1' ->
  eq_weak_heap H1 H1' ->
  strong_active (fst (fst (fst H1))) t ->
  (exists H2', exists P2',
    LazySys.system_step H1' P1' (Some (t, a)) H2' P2'
    /\ L.eq_program_state P2 P2'
    /\ eq_weak_heap H2 H2').
Proof with eauto.
  intros [ [ [ ts1 S1 ] R1 ] TS1 ] P1 [ [ [ ts1' S1' ] R1' ] TS1' ] P1' t a H2
    P2 WeakSysStep2 PrgmEq1 HeapEq1 StrongActive1.
  simpl in StrongActive1.
  inversion HeapEq1; subst.
  inversion WeakSysStep2 as [ ? ? ? ? ? ? ProgramStep2 WeakStep2 | ]; subst.
  forwards [ P2' [ PrgmEq2 ProgramStep2' ] ] : L.eq_program_state_step...
  inversion_weak_step WeakStep2 t' l r v v' WeakActive1 RgnFresh1 Fresh1 Read1
    TxFresh1 TxRead1 Unprotected1 Empty1;
    esplit; esplit; eauto 7.

  (* newRgn *)
  split. apply LazySys.system_step_mutual...
  apply Lazy.step_newRgn...
  rewrite <- VarMapFacts.In_m...
  split...

  (* alloc *)
  split. apply LazySys.system_step_mutual...
  apply Lazy.step_alloc...
  rewrite <- VarMapFacts.In_m...
  split...

  (* tx_read1 *)
  split. apply LazySys.system_step_mutual...
  eapply Lazy.step_tx_read1...
  rewrite <- VarMapFacts.MapsTo_m...
  rewrite <- VarMapFacts.In_m...
  split...

  (* tx_read2 *)
  split. apply LazySys.system_step_mutual...
  eapply Lazy.step_tx_read2...
  rewrite <- VarMapFacts.MapsTo_m...
  split...

  (* read *)
  inversion StrongActive1; subst.
  split. apply LazySys.system_step_mutual...
  eapply Lazy.step_read...
  rewrite <- VarMapFacts.MapsTo_m...
  split...
  inversion Unprotected1. tauto.

  (* write *)
  inversion StrongActive1; subst.
  split. apply LazySys.system_step_mutual...
  eapply Lazy.step_write...
  rewrite <- VarMapFacts.MapsTo_m...
  split...
  inversion Unprotected1. tauto.

  (* spawn *)
  inversion StrongActive1; subst...
  inversion Unprotected1. tauto.
Qed.

Lemma reordering_newRgn : forall H1 P1 e1 H2 P2 H2' P2' t r H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, newRgn _ r)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, newRgn _ r)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, newRgn _ r)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 9.
  intros [ [ [ ts1 S1 ] R1 ] TS1 ] P1 e1 [ [ [ ts2 S2 ] R2 ] TS2 ] P2
    [ [ [ ts2' S2' ] R2' ] TS2' ] P2' t r H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ProgramStep3 WeakStep3 | ]; subst.
  inversion_weak_step WeakStep3 t' l r' v v' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Unprotected2 Empty2.
  forwards [ P3' [ PrgmEq3 PrgmStep3' ] ] : L.eq_program_state_step...
  inversion Active2 as [ | t' | t' ]; subst;
    inversion HeapEq2 as [ ? ? ? ? ? ? ? StoreEq2 ProtEq2 TxStoreEq2 ]; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...

  (* running *)
  destruct (t == t').
  subst t'.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...

  (* what kind of lazy step? *)
  rewrite VarMapFacts.In_m in RgnFresh2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
  inversion_lazy_step LazyStep2' t'' l r' v v' Active1 RgnFresh1 Fresh1 Read1
  TxFresh1 TxRead1 Empty1;
  try (inversion Active1; subst);
    reorder_program_steps;
    forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
    eauto 17.

  (* newRgn/newRgn *)
  rewrite VarMapFactsEx.add_not_in_iff in RgnFresh2.
  destruct RgnFresh2.
  esplit. esplit. split. right. esplit. esplit. split...
  apply LazySys.system_step_mutual...
  apply Lazy.step_newRgn...
  rewrite VarMapFactsEx.add_not_in_iff...
  split... split...
  rewrite ProtEq2.
  apply VarMapFactsEx2.Equal_add_add...

  (* commit *)
  rewrite VarMapFacts.In_m in RgnFresh2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst.

  inversion_lazy_step LazyStep2' t'' l r' v v' Active1 RgnFresh1 Fresh1 Read1
    TxFresh1 TxRead1 Empty1; try inversion Active1;
      reorder_program_steps;
      forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
      eauto 16.

  inversion_lazy_step LazyStep2' t'' l r' v v' Active1 RgnFresh1 Fresh1 Read1
    TxFresh1 TxRead1 Empty1.

  (* for some reason eauto is not working on these cases... *)
  esplit. esplit. split. right. esplit. esplit. split...
  apply LazySys.system_step_heap.
  eapply Lazy.step_commit_ro...
  split...

  esplit. esplit. split. right. esplit. esplit. split...
  apply LazySys.system_step_heap.
  eapply Lazy.step_commit_mod...
  split...
Qed.

Lemma reordering_alloc : forall H1 P1 e1 H2 P2 H2' P2' t l r v H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, alloc l r v)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, alloc l r v)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, alloc l r v)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 9.
  intros [ [ [ ts1 S1 ] R1 ] TS1 ] P1 e1 [ [ [ ts2 S2 ] R2 ] TS2 ] P2
    [ [ [ ts2' S2' ] R2' ] TS2' ] P2' t l r v H3 P3 LazySysStep2' PrgmEq2
    HeapEq2 WeakSysStep3.
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ProgramStep3 WeakStep3 | ]; subst.
  inversion_weak_step WeakStep3 t' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Unprotected2 Empty2.
  forwards [ P3' [ PrgmEq3 PrgmStep3' ] ] : L.eq_program_state_step...
  inversion Active2 as [ | t' | t' ]; subst;
    inversion HeapEq2 as [ ? ? ? ? ? ? ? StoreEq2 ProtEq2 TxStoreEq2 ]; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...

  (* running *)
  destruct (t == t').
  subst t'.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...

  (* what kind of lazy step? *)
  rewrite VarMapFacts.In_m in Fresh2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
  inversion_lazy_step LazyStep2' t'' l' r' v' v'' Active1 RgnFresh1 Fresh1 Read1
  TxFresh1 TxRead1 Empty1;
  try (inversion Active1; subst);
    reorder_program_steps;
    forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
    eauto 17.

  (* alloc/alloc *)
  rewrite VarMapFactsEx.add_not_in_iff in Fresh2.
  destruct Fresh2.
  esplit. esplit. split. right. esplit. esplit. split...
  apply LazySys.system_step_mutual...
  apply Lazy.step_alloc...
  rewrite VarMapFactsEx.add_not_in_iff...
  split... split...
  rewrite StoreEq2.
  apply VarMapFactsEx2.Equal_add_add...

  (* tx_read1/alloc *)
  esplit. esplit. split. right. esplit. esplit. split...
  apply LazySys.system_step_mutual...
  eapply Lazy.step_tx_read1...
  rewrite VarMapFacts.add_neq_mapsto_iff...
  intro. subst. apply Fresh2...
  split...

  (* commit *)
  rewrite VarMapFacts.In_m in Fresh2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst.

  inversion_lazy_step LazyStep2' t'' l' r' v' v'' Active1 RgnFresh1 Fresh1 Read1
    TxFresh1 TxRead1 Empty1; try inversion Active1;
      reorder_program_steps;
      forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
      eauto 16.

  inversion_lazy_step LazyStep2' t'' l' r' v' v'' Active1 RgnFresh1 Fresh1 Read1
    TxFresh1 TxRead1 Empty1.

  (* for some reason eauto is not working on this case... *)
  esplit. esplit. split. right. esplit. esplit. split...
  apply LazySys.system_step_heap.
  eapply Lazy.step_commit_ro...
  split...

  (* commit_mod/alloc *)
  rewrite VarMapFactsEx.add_not_in_iff in Fresh2.
  destruct Fresh2...
  esplit. esplit. split. right. esplit. esplit. split.
  eapply WeakSys.system_step_mutual...
  apply LazySys.system_step_heap.
  eapply Lazy.step_commit_mod...
  rewrite VarMapFacts.add_neq_mapsto_iff...
  split... split...
  rewrite StoreEq2.
  apply VarMapFactsEx2.Equal_add_add...
Qed.

Lemma reordering_read : forall HT H1 P1 e1 H2 P2 H2' P2' t l v H3 P3,
  Typ.typing_prgm HT P1 ->
  TypedStrongLazy.typing_heap HT H1 ->
  weak_heap_invariant H1 ->
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  separation_weak H2 P2 ->
  WeakSys.system_step H2 P2 (Some (t, read l v)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, read l v)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, read l v)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros HT [ [ [ ts1 S1 ] R1 ] TS1 ] P1 e1 [ [ [ ts2 S2 ] R2 ] TS2 ] P2
    [ [ [ ts2' S2' ] R2' ] TS2' ] P2' t l v H3 P3 TypingPrgm1 TypingHeap1
    Invariant1 LazySysStep2' PrgmEq2 HeapEq2 SeparationWeak2 WeakSysStep3.
  inversion_typing_weak_heap TypingHeap1 DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ProgramStep3 WeakStep3 | ]; subst.
  forwards [ P3' [ PrgmEq3 PrgmStep3' ] ] : L.eq_program_state_step...
  inversion_weak_step WeakStep3 t' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Unprotected2 Empty2.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...
  inversion Unprotected2; subst;
    inversion HeapEq2 as [ ? ? ? ? ? ? ? StoreEq2 ProtEq2 TxStoreEq2 ]; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...

  (* running *)
  destruct (t == t').
  subst t'.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...

  (* what kind of lazy step? *)
  rewrite VarMapFacts.MapsTo_m in Read2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
  inversion_lazy_step LazyStep2' t'' l' r'' v' v'' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1;
  try (inversion Active1; subst);
    reorder_program_steps;
    forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
    eauto 17.

  (* read/alloc *)
  rewrite VarMapFacts.add_mapsto_iff in Read2.
  destruct Read2 as [ [ ? ? ] | [ ? ? ] ]. subst.
  contradiction Fresh1. rewrite <- DomStore.
  eapply LangReqs.well_formed_read...
  eauto 17.

  (* commit *)
  rewrite VarMapFacts.MapsTo_m in Read2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst.

  (* endP *)
  inversion_lazy_step LazyStep2' t'' l' r'' v' v'' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1; try inversion Active1;
    reorder_program_steps;
    forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
    eauto 16.

  (* empty step *)
  inversion_lazy_step LazyStep2' t'' l' r'' v' v'' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1.

  (* commit_ro/read *)
  eauto 17.

  (* commit_mod/read *)
  rewrite VarMapFacts.add_mapsto_iff in Read2.
  destruct Read2 as [ [ ? Eq ] | [ ? ? ] ].
  inversion Eq. subst.
  inversion SeparationWeak2 as [ | | ? ? ? ? ? ReadInactive ]; subst.
  specializes RgnStore (>>> Read1).
  rewrite DomProt in RgnStore.
  destruct RgnStore as [ p Prot1 ].
  forwards Check1 : ReadInactive...
  rewrite StoreEq2.
  rewrite VarMapFactsEx2.add_eq_mapsto_iff...
  rewrite ProtEq2...
  inversion Invariant1 as [ ? ? ? ? ? TxModState ]; subst.
  forwards Check2 : TxModState...
  inversion Check1; subst; inversion Check2. tauto.
  eauto 14.
Qed.

Lemma reordering_write : forall HT H1 P1 e1 H2 P2 H2' P2' t l v H3 P3,
  Typ.typing_prgm HT P1 ->
  TypedStrongLazy.typing_heap HT H1 ->
  weak_heap_invariant H1 ->
  separation_lazy H1 P1 ->
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  separation_weak H2 P2 ->
  WeakSys.system_step H2 P2 (Some (t, write l v)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, write l v)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, write l v)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros HT [ [ [ ts1 S1 ] R1 ] TS1 ] P1 e1 [ [ [ ts2 S2 ] R2 ] TS2 ] P2
    [ [ [ ts2' S2' ] R2' ] TS2' ] P2' t l v H3 P3 TypingPrgm1 TypingHeap1
    Invariant1 SeparationLazy1 LazySysStep2' PrgmEq2 HeapEq2 SeparationWeak2
    WeakSysStep3.
  inversion_typing_weak_heap TypingHeap1 DomStore DomTx TypingStore RgnStore
    TypingTx DomProt.
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ProgramStep3 WeakStep3 | ]; subst.
  forwards [ P3' [ PrgmEq3 PrgmStep3' ] ] : L.eq_program_state_step...
  inversion_weak_step WeakStep3 t' l' r' v' v'' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Unprotected2 Empty2.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...
  inversion Unprotected2; subst;
    inversion HeapEq2 as [ ? ? ? ? ? ? ? StoreEq2 ProtEq2 TxStoreEq2 ]; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...

  (* running *)
  destruct (t == t').
  subst t'.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
  simpl...

  (* what kind of lazy step? *)
  rewrite VarMapFacts.MapsTo_m in Read2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
  inversion_lazy_step LazyStep2' t'' l' r'' v'' v''' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1;
  try (inversion Active1; subst);
    reorder_program_steps;
    forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3').

  (* pure/write *)
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  split...

  (* newRgn/write *)
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  split...

  (* alloc/write *)
  rewrite VarMapFacts.add_mapsto_iff in Read2.
  destruct Read2 as [ [ ? ? ] | [ ? ? ] ]. subst.
  contradiction Fresh1. rewrite <- DomStore.
  eapply LangReqs.well_formed_write...
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  apply Lazy.step_alloc...
  rewrite VarMapFactsEx.add_not_in_iff...
  split... split...
  rewrite StoreEq2.
  apply VarMapFactsEx2.Equal_add_add...

  (* tx_read1/write *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  eapply Lazy.step_tx_read1...
  rewrite VarMapFacts.add_neq_mapsto_iff...
  intro. subst.
  forwards Eq : VarMapFacts.MapsTo_fun (>>> Read1 Read2).
  inversion Eq; subst.
  forwards Prot1 : RgnStore (>>> Read1).
  rewrite DomProt in Prot1.
  destruct Prot1 as [ p Prot1 ].
  inversion SeparationWeak2 as [ | ? ? ? ? ? ? WriteInactive | ]; subst.
  forwards Check1 : WriteInactive...
  rewrite StoreEq2...
  rewrite ProtEq2...
  inversion SeparationLazy1 as [ | ? ? ? ? ? ReadActive | ]; subst.
  forwards Check2 : ReadActive...
  inversion Check1; subst; inversion Check2. tauto.
  split...

  (* tx_read2/write *)
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  split...

  (* tx_write/write *)
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  split...

  (* beginP/write *)
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  split...

  (* commit *)
  rewrite VarMapFacts.MapsTo_m in Read2...
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst.

  (* endP *)
  inversion_lazy_step LazyStep2' t'' l' r'' v'' v''' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1; try inversion Active1.
  reorder_program_steps.
  forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3').
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_mutual...
  split...

  (* empty step *)
  inversion_lazy_step LazyStep2' t'' l' r'' v'' v''' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1.

  (* commit_ro/read *)
  (* This should work with eauto??? *)
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_heap...
  eapply Lazy.step_commit_ro...
  split...

  (* commit_mod/write *)
  rewrite VarMapFacts.add_mapsto_iff in Read2.
  destruct Read2 as [ [ ? Eq ] | [ ? ? ] ].
  inversion Eq. subst.
  inversion SeparationWeak2 as [ | | ? ? ? ? ? ? WriteInactive ]; subst.
  specializes RgnStore (>>> Read1).
  rewrite DomProt in RgnStore.
  destruct RgnStore as [ p Prot1 ].
  forwards Check1 : WriteInactive...
  rewrite StoreEq2...
  rewrite VarMapFactsEx2.add_eq_mapsto_iff...
  rewrite ProtEq2...
  inversion Invariant1 as [ ? ? ? ? ? TxModState ]; subst.
  forwards Check2 : TxModState...
  inversion Check1; subst; inversion Check2. tauto.
  esplit. esplit. split. right. esplit. esplit. split.
  apply WeakSys.system_step_mutual...
  eapply Weak.step_write...
  apply LazySys.system_step_heap...
  eapply Lazy.step_commit_mod...
  rewrite VarMapFacts.add_neq_mapsto_iff...
  split... split...
  rewrite StoreEq2.
  apply VarMapFactsEx2.Equal_add_add...
Qed.

Lemma reordering_spawn : forall H1 P1 e1 H2 P2 H2' P2' t t' H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, spawn _ t')) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, spawn _ t')) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, spawn _ t')) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros [ [ [ ts1 S1 ] R1 ] TS1 ] P1 e1 [ [ [ ts2 S2 ] R2 ] TS2 ] P2
    [ [ [ ts2' S2' ] R2' ] TS2' ] P2' t t' H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ProgramStep3 WeakStep3 | ]; subst.
  forwards [ P3' [ PrgmEq3 ProgramStep3' ] ] : L.eq_program_state_step...
  inversion_weak_step WeakStep3 t'' l r v v' Active2 RgnFresh2 Fresh2 Read2
    TxFresh2 TxRead2 Unprotected2 Empty2.
  inversion Unprotected2 as [ | ? t'' | ? t'' ]; subst;
    inversion HeapEq2; subst...

  (* running *)
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
  inversion_lazy_step LazyStep2' t''' l' r'' v'' v''' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1;
  try (inversion Active1; subst);
    reorder_program_steps;
    forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
    eauto 20.

  (* commit *)
  inversion LazySysStep2' as [ ? ? ? ? ? ? ProgramStep2' LazyStep2'
    | ? ? ? LazyStep2' ]; subst;
  inversion_lazy_step LazyStep2' t''' l' r'' v'' v''' Active1 RgnFresh1 Fresh1
  Read1 TxFresh1 TxRead1 Empty1;
  try (inversion Active1; subst);
    try reorder_program_steps;
      try forwards : L.eq_program_state_trans (>>> PrgmEq3 PrgmEq3');
        eauto 20.
Qed.

Lemma reordering_makeP : forall H1 P1 e1 H2 P2 H2' P2' t r H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, makeP _ r)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, makeP _ r)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, makeP _ r)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros H1 P1 e1 H2 P2 H2' P2' t r H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst.
  inversion HeapEq2; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
Qed.

Lemma reordering_makeU : forall H1 P1 e1 H2 P2 H2' P2' t r H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, makeU _ r)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, makeU _ r)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, makeU _ r)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros H1 P1 e1 H2 P2 H2' P2' t r H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst.
  inversion HeapEq2; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
Qed.

Lemma reordering_makeRO : forall H1 P1 e1 H2 P2 H2' P2' t r H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, makeRO _ r)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, makeRO _ r)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, makeRO _ r)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros H1 P1 e1 H2 P2 H2' P2' t r H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst.
  inversion HeapEq2; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
Qed.

Lemma reordering_makeTL : forall H1 P1 e1 H2 P2 H2' P2' t r H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, makeTL _ r)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, makeTL _ r)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, makeTL _ r)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros H1 P1 e1 H2 P2 H2' P2' t r H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst.
  inversion HeapEq2; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
Qed.

Lemma reordering_beginP : forall H1 P1 e1 H2 P2 H2' P2' t H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, beginP _)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, beginP _)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, beginP _)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros H1 P1 e1 H2 P2 H2' P2' t H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst.
  inversion HeapEq2; subst.
  forwards [ ? [ ? [ ? [ ? ? ] ] ] ] : reordering_strong_active_helper...
Qed.

Lemma reordering_endP : forall H1 P1 e1 H2 P2 H2' P2' t H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 (Some (t, endP _)) H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' (Some (t, endP _)) H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 (Some (t, endP _)) H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto.
  intros H1 P1 e1 H2 P2 H2' P2' t H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst.
  inversion HeapEq2; subst.
  forwards [ ? [ ? ? ] ] : L.eq_program_state_step...
  esplit. esplit. split...
Qed.

Lemma reordering_empty : forall H1 P1 e1 H2 P2 H2' P2' H3 P3,
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  WeakSys.system_step H2 P2 None H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' None H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 None H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto.
  intros H1 P1 e1 H2 P2 H2' P2' H3 P3 LazySysStep2' PrgmEq2 HeapEq2
    WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst.
  inversion WeakStep3; subst;
    inversion HeapEq2; subst.

  (* commit_ro *)
  esplit. esplit.
  split. left.
  apply LazySys.system_step_heap.
  eapply Lazy.step_commit_ro.
  rewrite <- VarMapFacts.MapsTo_m...
  split...
  apply eq_weak_heap_def...
  apply VarMapFacts.remove_m...

  (* commit_mod *)
  esplit. esplit.
  split. left.
  apply LazySys.system_step_heap.
  eapply Lazy.step_commit_mod.
  rewrite <- VarMapFacts.MapsTo_m...
  rewrite <- VarMapFacts.MapsTo_m...
  split...
  apply eq_weak_heap_def...
  apply VarMapFacts.remove_m...

  (* end_commit *)
  esplit. esplit.
  split. left.
  apply LazySys.system_step_heap.
  eapply Lazy.step_end_commit.
  intro. rewrite <- VarMapFacts.In_m...
  split...
Qed.

Lemma reordering : forall HT H1 P1 e1 H2 P2 H2' P2' e2 H3 P3,
  Typ.typing_prgm HT P1 ->
  TypedStrongLazy.typing_heap HT H1 ->
  weak_heap_invariant H1 ->
  separation_lazy H1 P1 ->
  LazySys.system_step H1 P1 e1 H2' P2' ->
  L.eq_program_state P2 P2' ->
  eq_weak_heap H2 H2' ->
  separation_weak H2 P2 ->
  WeakSys.system_step H2 P2 e2 H3 P3 ->
  (exists H3', exists P3',
    (LazySys.system_step H2' P2' e2 H3' P3'
      \/ (exists H2'', exists P2'',
        WeakSys.system_step H1 P1 e2 H2'' P2''
        /\ LazySys.system_step H2'' P2'' e1 H3' P3'))
    /\ L.eq_program_state P3 P3'
    /\ eq_weak_heap H3 H3').
Proof with eauto 10.
  intros HT H1 P1 e1 H2 P2 H2' P2' e2 H3 P3 TypingPrgm TypingHeap Invariant1
    SeparationLazy1 LazySysStep2' PrgmEq2 HeapEq2 SeparationWeak2 WeakSysStep3.
  destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
  inversion WeakSysStep3 as [ ? ? ? ? ? ? ? WeakStep3 | ? ? ? WeakStep3 ];
    subst; inversion WeakStep3; subst.
  eapply reordering_pure...
  eapply reordering_newRgn...
  eapply reordering_alloc...
  eapply reordering_read...
  eapply reordering_read...
  eapply reordering_read...
  eapply reordering_write...
  eapply reordering_write...
  eapply reordering_spawn...
  eapply reordering_makeP...
  eapply reordering_makeU...
  eapply reordering_makeRO...
  eapply reordering_makeTL...
  eapply reordering_beginP...
  eapply reordering_endP...
  eapply reordering_empty...
  eapply reordering_empty...
  eapply reordering_empty...
Qed.

Lemma weak_to_strong_lazy_helper : forall P0 H P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  separation P0 ->
  WeakSys.system_step_many empty_weak_heap P0 H P tr ->
  separation_weak H P
  /\ (exists H', exists P', exists tr',
    L.eq_program_state P P'
    /\ eq_weak_heap H H'
    /\ LazySys.system_step_many empty_weak_heap P0 H' P' tr'
    /\ length tr = length tr').
Proof with eauto.
  intros P0 H2 P2 tr2' TypingPrgm Separation0 WeakSysStepMany2.

  (* Proof by induction on the number of steps in the trace. *)
  remember (length tr2') as n.
  symmetry in Heqn.
  generalize dependent tr2'.
  generalize dependent P2.
  generalize dependent H2.
  induction n; intros H3 P3 tr3 WeakSysStepMany3 Heqn.

  (* n = 0 - nil trace *)
  rewrite <- length_nil in Heqn.
  subst.
  inversion WeakSysStepMany3; subst.
  split. apply separation_weak_empty...
  esplit. esplit. esplit.
  split. apply L.eq_program_state_refl.
  split. apply eq_weak_heap_refl.
  split. apply LazySys.system_step_many_refl.
  auto.

  (* length tr = S n - trace has at least one step. *)
  apply length_cons in Heqn.
  destruct Heqn as [ e3 [ tr2 [ ? ? ] ] ]; subst.
  inversion WeakSysStepMany3 as [ | ? ? H2 P2 ? ? ? ? WeakSysStepMany2
    WeakSysStep3 ]; subst.

  (* Apply the inductive hypothesis to get a lazy execution for the
     first n steps. *)
  forwards [ SeparationWeak2 [ H2' [ P2' [ tr2' [ PrgmEq2 [ HeapEq2
    [ LazySysStepMany2' LengthEq2 ] ] ] ] ] ] ] : IHn (>>> WeakSysStepMany2)...
  inversion LazySysStepMany2' as [ | ? ? H1 P1 ? ? tr1 e2 LazySysStepMany1
    LazySysStep2' ]; subst.

  (* If the execution has zero steps, then the weak step started from
     the tx_none state, which means it's also OK in the lazy
     semantics. The weak and lazy semantics only differ when starting
     from the active or commit states. *)
  rewrite <- length_nil in LengthEq2. subst.
  inversion WeakSysStepMany2; subst.
  forwards LazySysStep3 : weak_to_strong_lazy_none (>>> WeakSysStep3).
  split. eapply separation_lazy_to_weak...
  apply TypedStrongLazy.typing_empty_heap.
  eapply separation_to_separation_lazy...
  eapply LazySys.system_step_many_trans...
  esplit. esplit. esplit.
  split. apply L.eq_program_state_refl.
  split. apply eq_weak_heap_refl.
  split. eapply LazySys.system_step_many_trans...
  auto.

  (* Otherwise, the lazy execution had at least one step, so we can apply the
     reordering lemma. *)
  assert (separation_lazy empty_weak_heap P0) as SeparationLazy0.
  eapply separation_to_separation_lazy...
  forwards SeparationLazy1 : separation_to_separation_lazy
    (>>> LazySysStepMany1)...
  forwards SeparationLazy2' : separation_to_separation_lazy
    (>>> LazySysStepMany2')...
  forwards [ HT1 [ TypingHeap1 TypingPrgm1 ] ] :
    LazySysPreservation.type_preservation_many (>>> LazySysStepMany1)...
  forwards [ HT2 [ TypingHeap2' TypingPrgm2' ] ] :
    LazySysPreservation.type_preservation_many (>>> LazySysStepMany2')...
  forwards Invariant1 : weak_heap_invariant_step_many (>>> LazySysStepMany1)...
  forwards Invariant2' : weak_heap_invariant_step_many
    (>>> LazySysStepMany2')...
  pose proof reordering TypingPrgm1 TypingHeap1 Invariant1 SeparationLazy1
    LazySysStep2' PrgmEq2 HeapEq2 SeparationWeak2 WeakSysStep3
    as [ H3' [ P3' [ [ LazySysStep3' | [ H2'' [ P2'' [ WeakSysStep2''
      LazySysStep3' ] ] ] ] [ PrgmEq3 HeapEq3 ] ] ] ].

  (* The step taken by the weak semantics is allowed in the lazy semantics. *)
  split. eapply separation_weak_eq.
  eapply separation_lazy_to_weak.
  apply TypingPrgm2'. apply TypingHeap2'. auto.
  eapply separation_weak_eq.
  apply SeparationWeak2. auto. auto. eauto.
  eapply separation_to_separation_lazy...
  eapply LazySys.system_step_many_trans...
  apply eq_weak_heap_sym...
  apply L.eq_program_state_sym...
  esplit. esplit. esplit.
  split... split... split.
  eapply LazySys.system_step_many_trans...
  simpl...

  (* We can reorder the lazy and weak steps and apply the inductive
     hypothesis. *)
  forwards [ SeparationWeak2'' [ H2''' [ P2''' [ tr2' [ PrgmEq2' [ HeapEq2'
    [ LazySysStepMany2''' ? ] ] ] ] ] ] ] : IHn.
  eapply WeakSys.system_step_many_trans.
  apply strong_lazy_to_weak.
  apply LazySysStepMany1.
  apply WeakSysStep2''.
  auto.

  (* And now we can conclude. *)
  forwards [ H3'' [ P3'' [ ? [ ? ? ] ] ] ] :
  system_step_eq (>>> LazySysStep3')...
  split. apply separation_weak_eq with (H := H3'') (P := P3'').
  eapply separation_lazy_to_weak_many...
  eapply LazySys.system_step_many_trans...
  eapply eq_weak_heap_trans; apply eq_weak_heap_sym...
  eapply L.eq_program_state_trans; apply L.eq_program_state_sym...
  exists H3''. exists P3''. esplit.
  split. eapply L.eq_program_state_trans...
  split. eapply eq_weak_heap_trans...
  split. eapply LazySys.system_step_many_trans...
  simpl...
Qed.

Theorem weak_to_strong_lazy : forall P0 H P tr,
  Typ.typing_prgm (HeapTyping.empty_heap_typing _) P0 ->
  separation P0 ->
  WeakSys.system_step_many Weak.empty_heap P0 H P tr ->
  exists H', exists P', exists tr',
    L.eq_program_state P P'
    /\ LazySys.system_step_many Lazy.empty_heap P0 H' P' tr'.
Proof with eauto.
  intros P0 H P tr TypingPrgm Separation StepMany.
  forwards [ ? [ ? [ ? [ ? [ ? [ ? [ ? ? ] ] ] ] ] ] ]
    : weak_to_strong_lazy_helper...
Qed.
